import decimal
import itertools
import os
import pathlib
import re

import canmatrix
import click.testing
import pytest

import epyqlib.tests.common

import epcpm.cli.importsym
import epcpm.cli.exportsym
import epcpm.importexport
import epcpm.importexportdialog


round_trip = pathlib.Path(__file__).parents[3] / 'roundtrip'
project = round_trip / 'project.pmp'
example_device = epyqlib.tests.common.new_devices()[('develop', 'factory')]
exported_can = (
        round_trip / 'exported'
).with_suffix(example_device.can.suffix)
exported_hierarchy = (
        round_trip / 'hierarchy.parameters.json'
).with_suffix(example_device.hierarchy.suffix)


def _full_round_trip():
    # directory = pathlib.Path('full_import')
    # paths = epcpm.importexportdialog.paths_from_directory(directory)
    # project = epcpm.importexport.full_import(paths=paths)
    import epcpm.project
    project = epcpm.project.loadp(pathlib.Path('full_import')/'project.pmp')

    directory = pathlib.Path(os.sep)/'epc'/'g'/'36'/'grid-tied'
    paths = epcpm.importexportdialog.paths_from_directory(directory)

    epcpm.importexport.full_export(project=project, paths=paths, first_time=True)

    # project = epcpm.importexport.full_import(paths=paths)
    # epcpm.importexport.full_export(project=project, paths=paths)


@pytest.mark.skip
def test_import():
    runner = click.testing.CliRunner()

    result = runner.invoke(
        epcpm.cli.importsym.cli,
        [
            '--sym', example_device.can,
            '--hierarchy', example_device.hierarchy,
            '--project', project,
            '--can', round_trip / 'can.json',
            '--parameters', round_trip / 'parameters.json',
            '--sunspec', round_trip / 'sunspec.json',
            '--add-tables',
            '--add-sunspec-types',
        ],
        catch_exceptions=False,
    )
    print(result.output)

    assert result.exit_code == 0


@pytest.mark.skip
def test_export():
    runner = click.testing.CliRunner()

    result = runner.invoke(
        epcpm.cli.exportsym.cli,
        [
            '--project', project,
            '--sym', exported_can,
            '--hierarchy', exported_hierarchy,
        ],
        catch_exceptions=False,
    )
    print(result.output)

    assert result.exit_code == 0


def normalized_attributes(attributes):
    attributes = dict(attributes)

    return attributes


def assert_signals_equal(
        original,
        original_frame,
        exported,
        exported_frame,
):
    print(
        original_frame.name,
        original_frame.mux_names.get(original.multiplex),
        original.name,
    )
    assert_attributes_equal(
        original=original,
        exported=exported,
        attributes=(
            'enumeration',
            'is_float',
            'is_little_endian',
            'is_multiplexer',
            'is_signed',
            'max',
            'min',
            'multiplex',
            'factor',
            'offset',
            'size',
            'startBit',
            'unit',
        ),
    )

    tweaked = dict(exported.attributes)
    if 'LongName' not in original.attributes:
        tweaked.pop('LongName', None)

    tweaked_attributes = normalized_attributes(tweaked)
    original_attributes = normalized_attributes(original.attributes)

    assert tweaked_attributes == original_attributes, original.name

    factory = '<factory>'
    configs = ('<HY>', '<MG3>', '<MG4>', '<DG>', '<DC>')

    def split_and_access_level(comment):

        split = comment.split()

        return (
            factory in split,
            (
                tag in split
                for tag in configs
            ),
            [x for x in split if x not in (factory, *configs)],
        )

    original_factory, original_configs, original_comment = (
        split_and_access_level(original.comment)
    )
    exported_factory, exported_configs, exported_comment = (
        split_and_access_level(exported.comment)
    )

    assert exported_comment == original_comment, original.name

    not_applicable = (
        'readparam_command',
        'readparam_status',
        'meta',
    )
    if original.name.casefold() not in not_applicable:
        original_mux = [
            x for x in original_frame.signals if x.multiplex == 'Multiplexor'
        ]
        original_mux, = original_mux if original_mux else (None,)

        exported_mux = [
            x for x in exported_frame.signals if x.multiplex == 'Multiplexor'
        ]
        exported_mux, = exported_mux if exported_mux else (None,)

        if original_mux is None:
            assert exported_mux == original_mux, original.name

        def tagged_including_mux_check(tag, tagged, signal, mux):
            if mux is not None:
                return (
                        tagged
                        or (
                                signal is not mux
                                and tag in mux.comments[
                                    signal.multiplex]
                        )
                )

            return (
                    tagged
                    or tag in original_frame.comment
            )

        original_access_level = tagged_including_mux_check(
            tag=factory,
            tagged=original_factory,
            signal=original,
            mux=original_mux,
        )
        exported_access_level = tagged_including_mux_check(
            tag=factory,
            tagged=exported_factory,
            signal=exported,
            mux=exported_mux,
        )

        assert exported_access_level == original_access_level, original.name

        z = zip(configs, original_configs, exported_configs)
        for config, original_tagged, exported_tagged in z:
            original_config = tagged_including_mux_check(
                tag=config,
                tagged=original_tagged,
                signal=original,
                mux=original_mux,
            )
            exported_config = tagged_including_mux_check(
                tag=config,
                tagged=exported_tagged,
                signal=exported,
                mux=exported_mux,
            )

            assert exported_config == original_config, (config, exported.comment, original.comment, original.name)

    assert_varied_dict_equal(original.values, exported.values)

    assert exported_comment == original_comment, original.name


def get_name(x):
    return x.name


def sorted_names(names):
    return sorted(names, key=lambda x: (x.casefold(), x))


def unvary_signal_names(name):
    changes = (
        (r'(IGBT|Diode)_R(\d)', r'\1_r\2'),
        (r'(IGBT|Diode)_Tc(\d)', r'\1_tc\2'),
        ('Rii', 'rii'),
        ('Rdi', 'rdi'),
        ('Alpha', 'alpha'),
        ('Amps', 'amps'),
        ('Beta', 'beta'),
        (r'Leg(\d)', r'leg\1'),
        ('CurrentLCL', 'currentLCL'),
        ('MinPulseTime', 'minPulseTime'),
        ('N15V_Supply', 'n15V_Supply'),
        (r'V([p\d]+)_Supply', r'v\1_Supply'),
        ('Phase([ABC])', r'phase\1'),
        ('PostTriggerDuration_S', 'PostTriggerDuration_s'),
        (r'LegInductance(\d)', r'legInductance\1'),
        ('TempSetpoint', 'tempSetpoint'),
        ('LineCurrentLimit', 'lineCurrentLimit'),
        ('Dc(Current|Voltage)Limit', r'dc\1Limit'),
        ('Dc(Current|Voltage)Limit_Echo', r'dc\1Limit_echo'),
        ('Line(Current|Voltage)Limit_Echo', r'line\1Limit_echo'),
        ('PpcIslDroop([VF])_IBpt', r'PpcIslDroop\1_iBpt'),
        ('PpcIslDroop([VF])_DBpt', r'PpcIslDroop\1_dBpt'),
        ('PpcIslDroop([VF])_QBpt', r'PpcIslDroop\1_qBpt'),
        ('PpcIslDestab([VF])_VBpt', r'PpcIslDestab\1_vBpt'),
        ('PpcIslDestab([VF])_DBpt', r'PpcIslDestab\1_dBpt'),
        ('PpcIslDestab([VF])_QBpt', r'PpcIslDestab\1_qBpt'),
        ('PpcIslDestab([VF])_FBpt', r'PpcIslDestab\1_fBpt'),
        ('Ig_Max', 'Ig_max'),
        (
            r'(MesdConfigLutCalibration_)T(Igbt|Aux|Motor)_(\d+)',
            r'\1t\2_\3',
        ),
        (
            '(ProtTemperatureLimitTrip_)T({})_(Running|Stopped)'.format(
                '|'.join((
                    'Inverter',
                    'Internal',
                    'Device',
                    'Inlet',
                    'Delta',
                )),
            ),
            lambda m: f'{m.group(1)}t{m.group(2)}_{m.group(3).casefold()}',
        ),
        (
            '(ProtCtrlVoltageLimitTrip)_V(Pos|Neg)(24|15|5|3)',
            r'\1_v\2\3',
        ),
        (r'(.*)_TxRate', r'\1_txRate'),
        *(
            (x, lambda m: m.group(0)[0].casefold() + m.group(0)[1:])
            for x in (
                'DIg',
                'DpwmStrategy',
                'DcGroup',
                'D[012]',
                'F[012]',
                'Q([012]|Ig)',
                'V([012])',
                '(F|V|W|XpVa)Tau',
                'AcGroup',
                '(C|G)(L|R)s',
                'AcHalfPeriodTrack',
                'DtCompEnable',
                'DutyLimit',
                'EeSaveInProgress',
                'EntryPulseNr',
                'FCut(F|Ig|V|Vg)',
                'FDev(H|L)',
                'FMax',
                'FMin',
                'FNom',
                'FourWireEnable',
                'FSw(Hysteresis|Minimum|Request)',
                'I([012]|Abs|Delta|Dlt)',
                'Ind(IslStart|Threshold)',
                'K(Alpha|Apply|Phi|Vq)',
                'Limit(XY|Z)',
                'Lim(Lower|Upper)',
                'Load(L|PhiTau|R)',
                'Min(Current|VectorTime|Volt)',
                'ModDepthLimit(Beg|End)',
                'NetworkFrequency',
                'P(Delta|PhaseSel)',
                'PhaseSel',
                'PhiAbs',
                'Sat(CompEnable|Voltage)',
                'S(C|R)s',
                'SRv',
                'SimModeEnable',
                'SpectrumAdjustTau',
                'SpreadSpect(Enable|Inject)',
                'StandardDeviation',
                'SyncMod(Enable|Strategy)',
                'Tau',
                'T(Adapt|Deadband|Delta)',
                'Threshold',
                'TimeConst',
                'TrigCalTime',
                'TripTime',
                'TSwitch',
                'U',
                'VDelta',
                'VectAlignCtrl_(K[dp]|StepSat|Ti)',
                'V(Max|Min)',
                'VoltMag',
                'VsComp(Enable|Tau)',
                'Y',
                'ZeroVectorStrategy',
            )
        ),
        *(
            (
                r'(.*)_({x})_(.*)'.format(x=x),
                lambda
                    m: f'{m.group(1)}_{m.group(2).casefold()}_{m.group(3)}',
            )
            for x in (
                'Seconds',
                'Hertz',
                'Percent_Nominal_Volts',
                'Percent_Nominal_Var',
                'Percent_Nominal_Pwr',
                'Percent',
                'Nominal_Volts',
            )
        ),
        *(
            (x, lambda m: m.group(1) + m.group(2) + m.group(3).casefold())
            for x in (
                '(DcCurrent|DcVoltage|ReactivePower|RealPower)()(Limit|_Echo)',
                '(ReadParam|SaveToEE)(_)(Command|Status)',
                '(TempTmodel)(_)(Calculated)',
                '({})(_)(Echo)'.format(
                    '|'.join((
                        'Enable',
                        'FaultClr',
                        'WarningClr',
                        'ReactiveCurrent',
                        'RealCurrent',
                        'IslandReconnect',
                    )),
                ),
                '({})(_)(Measured)'.format(
                    '|'.join((
                        'TempDeviceBrdg[ABC]',
                        'CurrentAC',
                        'Frequency',
                        'VoltageAC',
                        'FrequencyRemote',
                        'VoltageAcRemote',
                        'SolarCurrent',
                        'SolarVoltage',
                        'CurrentDC',
                        'VoltageDCBus',
                        'VoltageDCInput',
                        'L[123]Current',
                        'L[123]ToL[123]Voltage',
                        '.*',
                    )),
                ),
        )
        ),
    )

    tweaked = name
    for pattern, replacement in changes:
        tweaked = re.sub(f'^{pattern}$', replacement, tweaked)

    return tweaked


def assert_signal_list_equal(
        original,
        original_frame,
        exported,
        exported_frame,
):
    original_names = sorted_names(
        x.name
        for x in original
    )
    exported_names = sorted_names(
        unvary_signal_names(x.name)
        for x in exported
    )

    case_insensitive = (
        'aclosshotstandbydc',
        'aclosshotstandbypc',
        'acwaitenabledcatref',
        'argadjust',
        'argerrmax',
        'argerrsat',
        'argerrtau',
        'autodisconnectongridloss',
        'autofollowingpostconnect',
        'autoreconnectongridavail',
        'autosequencewhendisabled',
        'autovolttrackduringsynch',
        'conntimeout',
        'controlsvariant',
        'current',
        'dcwaitenabledcatref',
        'disctimeout',
        'enabledelayac',
        'enabledelaydc',
        'didt',
        'dqccterm',
        'dummy',
        'freq',
        'fswactual',
        # 'ftau',
        'gain',
        'groupedenable',
        'hardwarevariant',
        'level',
        'limit',
        'magadjust',
        'magerrmax',
        'magerrtau',
        'moddepthxy',
        'moddepthz',
        'max',
        'min',
        'pllenable',
        'pulsenractual',
        'r',
        'rate',
        'resistance',
        'syncactive',
        'syncholdtime',
        'synctimeout',
        'tc',
        'temp',
        'temperature',
        'time',
        'tisllatch',
        'tunload',
        'vdevh',
        'vdevl',
        'volts',
        # 'vtau',
        # 'wtau',
    )

    zipped = itertools.zip_longest(original_names, exported_names)
    for original_name, exported_name in zipped:
        if original_name.casefold() in case_insensitive:
            assert (
                exported_name.casefold()
                == original_name.casefold()
            ), original.name
        else:
            assert exported_name == original_name

    def key(x):
        return (
            x.name.casefold(), x.multiplex
            if x.multiplex is not None
            else float('-inf')
        )

    zipped = itertools.zip_longest(
        sorted(original, key=key),
        sorted(exported, key=key),
    )
    for original_signal, exported_signal in zipped:
        assert_signals_equal(
            original=original_signal,
            original_frame=original_frame,
            exported=exported_signal,
            exported_frame=exported_frame,
        )


def assert_attributes_equal(original, exported, *, attributes):
    for attribute in attributes:
        assert (
            getattr(exported, attribute) == getattr(original, attribute)
        ), (original.name, attribute)


def assert_varied_dict_equal(original, exported):
    zipped = itertools.zip_longest(
        sorted(original.items()),
        sorted(exported.items()),
    )
    for (original_value, original_name), (exported_value, exported_name) in zipped:
        assert exported_value == original_value

        tweaked = unvary_signal_names(exported_name)
        assert tweaked == original_name


def assert_frames_equal(original, exported):
    assert_attributes_equal(
        original=original,
        exported=exported,
        attributes=(
            'name',
            'id',
            'extended',
            'attributes',
            'comment',
            'is_multiplexed',
            'size',
        ),
    )

    assert_varied_dict_equal(original.mux_names, exported.mux_names)

    assert_signal_list_equal(
        original=original.signals,
        original_frame=original,
        exported=exported.signals,
        exported_frame=exported,
    )


def collect_attribute(sequence, name):
    return {getattr(x, name) for x in sequence}


def assert_frame_lists_equal(original, exported):
    original_names = collect_attribute(original, 'name')
    exported_names = collect_attribute(exported, 'name')

    assert exported_names == original_names, original.name

    zipped = itertools.zip_longest(
        sorted(original, key=get_name),
        sorted(exported, key=get_name),
    )
    for original_frame, exported_frame in zipped:
        assert_frames_equal(original_frame, exported_frame)


def assert_value_tables_equal(original, exported):
    return exported == original


@pytest.mark.skip
def test_roundtrip():
    original, = canmatrix.formats.loadp(os.fspath(example_device.can)).values()
    assert original.load_errors == []
    exported, = canmatrix.formats.loadp(os.fspath(exported_can)).values()
    assert exported.load_errors == []

    assert_frame_lists_equal(original.frames, exported.frames)
    assert_value_tables_equal(original.value_tables, exported.value_tables)



# This file has been generated


build_system = 'AppVeyor'
build_id = '33297084'
build_number = '663'
build_version = '1.0.663'
job_id = 'dmir2rlxs60ywhy0'
job_url = 'https://ci.appveyor.com/project/KyleAltendorf/pm/build/job/dmir2rlxs60ywhy0'



import pathlib

import click
import graham

import epyqlib.attrsmodel
import epyqlib.pm.parametermodel
import epyqlib.pm.valuesetmodel

import epcpm.canmodel
import epcpm.project
import epcpm.symtoproject


def relative_path(target, reference):
    reference = pathlib.Path(reference)
    if reference.is_file():
        reference = reference.parents[0]

    return pathlib.Path(target).resolve().relative_to(reference.resolve())


@click.command()
@click.option('--sym', type=click.File('rb'), required=True)
@click.option('--hierarchy', type=click.File(), required=True)
@click.option('--project', type=click.File('w'), required=True)
@click.option('--parameters', type=click.File('w'), required=True)
@click.option('--can', type=click.File('w'), required=True)
@click.option('--sunspec', type=click.File('w'), required=True)
@click.option('--epyq-value-set', type=click.File('w'))
@click.option('--add-tables/--no-add-tables', default=False)
@click.option('--add-sunspec-types/--no-add-sunspec-types', default=False)
def cli(
        sym,
        hierarchy,
        project,
        parameters,
        can,
        sunspec,
        epyq_value_set,
        add_tables,
        add_sunspec_types,
):
    parameters_root, can_root, sunspec_root = epcpm.symtoproject.load_can_file(
        can_file=sym,
        file_type=str(pathlib.Path(sym.name).suffix[1:]),
        parameter_hierarchy_file=hierarchy,
    )

    project_path = pathlib.Path(project.name).parents[0]
    project_path.mkdir(parents=True, exist_ok=True)

    project_model = epcpm.project.Project(
        paths=epcpm.project.Models(
            parameters=relative_path(parameters.name, project_path),
            can=relative_path(can.name, project_path),
            sunspec=relative_path(sunspec.name, project_path),
        ),
        models=epcpm.project.Models(
            parameters=epyqlib.attrsmodel.Model(
                root=parameters_root,
                columns=epyqlib.pm.parametermodel.columns,
            ),
            can=epyqlib.attrsmodel.Model(
                root=can_root,
                columns=epcpm.canmodel.columns,
            ),
            sunspec=epyqlib.attrsmodel.Model(
                root=sunspec_root,
                columns=epcpm.sunspecmodel.columns,
            ),
        ),
    )

    if add_sunspec_types:
        parameter_root = project_model.models.parameters.root
        enumerations = parameter_root.child_by_name('Enumerations')
        enumerations.append_child(
            epcpm.sunspecmodel.build_sunspec_types_enumeration(),
        )

    epcpm.project._post_load(project_model)

    if add_tables:
        epcpm.symtoproject.go_add_tables(
            parameters_root=project_model.models.parameters.root,
            can_root=project_model.models.can.root,
        )

    project.write(graham.dumps(project_model, indent=4).data)
    parameters.write(graham.dumps(parameters_root, indent=4).data)
    can.write(graham.dumps(can_root, indent=4).data)
    sunspec.write(graham.dumps(sunspec_root, indent=4).data)

    if epyq_value_set is not None:
        value_set = epyqlib.pm.valuesetmodel.create_blank(
            parameter_model=project_model.models.parameters,
        )

        parameters_root, = [
            node
            for node in value_set.parameter_model.root.children
            if node.name == 'Parameters'
        ]

        epyqlib.pm.valuesetmodel.copy_parameter_data(
            value_set=value_set,
            human_names=False,
            base_node=parameters_root,
            calculate_unspecified_min_max=True,
            symbol_root=can_root,
        )

        epyq_value_set.write(graham.dumps(value_set.model.root, indent=4).data)



import attr

import epcpm.c
import epcpm.sunspecmodel
import epcpm.sunspectoxlsx
import epyqlib.utils.general


builders = epyqlib.utils.general.TypeMap()


def export(path, sunspec_model):
    builder = builders.wrap(
        wrapped=sunspec_model.root,
        parameter_uuid_finder=sunspec_model.node_from_uuid,
        path=path,
    )

    path.mkdir(parents=True, exist_ok=True)
    builder.gen()


def gen(self, first=0):
    lines = []

    for member in self.wrapped.children[first:]:
        builder = builders.wrap(
            wrapped=member,
            parameter_uuid_finder=self.parameter_uuid_finder,
        )
        lines.extend(builder.gen())
        lines.append('')

    return lines


@builders(epcpm.sunspecmodel.Root)
@attr.s
class Root:
    wrapped = attr.ib()
    parameter_uuid_finder = attr.ib()
    path = attr.ib()

    def gen(self):
        for member in self.wrapped.children:
            builder = builders.wrap(
                wrapped=member,
                parameter_uuid_finder=self.parameter_uuid_finder,
            )
            path = self.path/f'sunspecInterface{member.id:05}.c'
            lines = [
                '#include "faultHandler.h"',
                '#include "measurementd.h"',
                '#include "phaseControl.h"',
                '#include "referenceHandler.h"',
                # '#include "hardware.h"',
                # '#include "modbusHandler.h"',
                # '#include "uart.h"',
                '',
                '#include "sunspecInterfaceGen.h"',
                '#include "{}"'.format(path.with_suffix('.h').name),
                f'#include "sunspecModel{member.id}.h"',
                '',
                '',
            ]
            lines.extend(builder.gen())
            with path.open('w', newline='\n') as f:
                f.write(epcpm.c.format_nested_lists(lines).strip() + '\n')


@builders(epcpm.sunspecmodel.Model)
@attr.s
class SunSpecModel:
    wrapped = attr.ib()
    parameter_uuid_finder = attr.ib()

    def gen(self):
        return gen(self=self, first=1)


@builders(epcpm.sunspecmodel.HeaderBlock)
@builders(epcpm.sunspecmodel.FixedBlock)
@attr.s
class Block:
    wrapped = attr.ib()
    parameter_uuid_finder = attr.ib()

    gen = gen


@builders(epcpm.sunspecmodel.DataPoint)
@attr.s
class DataPoint:
    wrapped = attr.ib()
    parameter_uuid_finder = attr.ib()

    def gen(self):
        if self.wrapped.parameter_uuid is None:
            return []

        parameter = self.parameter_uuid_finder(self.wrapped.parameter_uuid)

        get_name = epcpm.sunspectoxlsx.getter_setter_name(
            get_set='get',
            point=self.wrapped,
            parameter=parameter,
        )

        set_name = epcpm.sunspectoxlsx.getter_setter_name(
            get_set='set',
            point=self.wrapped,
            parameter=parameter,
        )

        lines = []

        lines.extend([
            f'void {get_name}(void) {{',
            *(
                [self.wrapped.get.splitlines()]
                if self.wrapped.get is not None
                else []
            ),
            '}',
            '',
            f'void {set_name}(void) {{',
            *(
                [self.wrapped.set.splitlines()]
                if self.wrapped.set is not None
                else []
            ),
            '}',
        ])

        return lines



# import collections
import pathlib

import pytest
import sunspec.core.client

import epcpm.smdxtosunspec


this = pathlib.Path(__file__).resolve()
here = this.parent


smdx_path = here/'sunspec'


@pytest.fixture
def fresh_pathlist():
    original_pathlist = sunspec.core.device.file_pathlist
    sunspec.core.device.file_pathlist = sunspec.core.util.PathList()

    yield sunspec.core.device.file_pathlist

    sunspec.core.device.file_pathlist = original_pathlist


def test_load():
    project = epcpm.project.loadp(here/'project'/'project.pmp')

    parameter_model = project.models.parameters
    enumerations = parameter_model.list_selection_roots['enumerations']
    sunspec_types = epcpm.sunspecmodel.build_sunspec_types_enumeration()
    enumerations.append_child(sunspec_types)
    parameter_model.list_selection_roots['sunspec types'] = sunspec_types

    requested_models = [1, 17, 103, 65534]

    models = epcpm.smdxtosunspec.import_models(
        *requested_models,
        parameter_model=project.models.parameters,
        paths=[smdx_path],
    )

    assert [model.id for model in models] == requested_models



import pathlib
import textwrap

import epyqlib.attrsmodel
import epyqlib.pm.parametermodel

import epcpm.canmodel
import epcpm.cantosym
import epcpm.project


here = pathlib.Path(__file__).parent


# def test_explore():
#     root = epcpm.canmodel.Root()
#
#     message = epcpm.canmodel.Message(
#         name='Test Message',
#     )
#     root.append_child(message)
#
#     signal = epcpm.canmodel.Signal(
#         name='Test Signal',
#         signed=True,
#     )
#     message.append_child(signal)
#
#     builder = epcpm.cantosym.builders.wrap(root)
#     s = builder.gen()
#
#     assert s == textwrap.dedent('''\
#     FormatVersion=5.0 // Do not edit this line!
#     Title="canmatrix-Export"
#     {ENUMS}
#
#
#     {SENDRECEIVE}
#
#     [TestMessage]
#     ID=1FFFFFFFh
#     Type=Extended
#     DLC=0
#     Var=TestSignal signed 0,0
#
#     ''')


def try_loading_multiplex():
    import io
    f = io.BytesIO(textwrap.dedent('''\
    {SEND}
    
    [TestMultiplexedMessage]
    ID=1FFFFFFFh
    Type=Extended
    DLC=0
    Mux=MuxA 0,8 0
    Var=SignalA unsigned 0,0
    
    [TestMultiplexedMessage]
    DLC=0
    Mux=MuxB 0,8 1
    Var=SignalB signed 0,0
    ''').encode('utf-8'))
    import canmatrix
    matrix, = canmatrix.formats.load(
        fileObject=f,
        importType='sym',
    ).values()

    print()


def tidy_sym(s):
    return '\n'.join(s.strip() for s in s.splitlines()).strip()


def test_multiplexed():
    root = epcpm.canmodel.Root()

    parameter_root = epyqlib.pm.parametermodel.Root(
        uuid='5ad34154-f585-40bc-992a-0c9685cbdb26',
    )
    model = epyqlib.attrsmodel.Model(
        root=parameter_root,
        columns=epyqlib.pm.parametermodel.columns,
    )
    model.add_drop_sources(model)
    access_levels = epyqlib.pm.parametermodel.AccessLevels(
        uuid='284090fd-6893-4800-92bc-035dbf25909c',
    )
    access_level = epyqlib.pm.parametermodel.AccessLevel(
        value=0,
        uuid='40ced595-a5f3-41ee-9156-eb0ac343c9a5',
    )
    access_levels.append_child(access_level)
    enumerations = epyqlib.pm.parametermodel.Enumerations(
        uuid='af9bfcf3-1858-4fb9-b238-a7641bcffcd4',
    )
    enumerations.append_child(access_levels)
    parameter_root.append_child(enumerations)
    builder = epcpm.cantosym.builders.wrap(
        wrapped=root,
        parameter_uuid_finder=model.node_from_uuid,
        parameter_model=model,
        access_levels=access_levels,
    )

    multiplexed_message = epcpm.canmodel.MultiplexedMessage(
        name='TestMultiplexedMessage',
        identifier=0x1abeface,
    )
    root.append_child(multiplexed_message)

    expected = textwrap.dedent('''\
    FormatVersion=5.0 // Do not edit this line!
    Title="canmatrix-Export"
    {ENUMS}


    {SENDRECEIVE}

    [TestMultiplexedMessage]
    ID=1ABEFACEh
    Type=Extended
    DLC=0
    ''')

    # assert tidy_sym(builder.gen()) == tidy_sym(expected)

    multiplex_signal = epcpm.canmodel.Signal(
        name='MultiplexerSignal',
        bits=8,
        uuid='c812b3be-c883-4828-8695-0f61db24fde3',
    )
    multiplexed_message.append_child(multiplex_signal)
    common_signal = epcpm.canmodel.Signal(
        name='CommonSignal',
        signed=True,
        uuid='3ecc047e-e88c-4917-b22c-2751db6db705',
    )
    multiplexed_message.append_child(common_signal)

    # seems like at this point it should do the same thing, or perhaps
    # add a line for the multiplexer signal.  instead it wipes the message
    # out entirely.
    # expected += textwrap.dedent('''\
    # Mux=MultiplexerSignal0 0,8 0
    # Var=CommonSignal signed 0,0
    # ''')
    # assert tidy_sym(builder.gen()) == tidy_sym(expected)

    parameter_a = epyqlib.pm.parametermodel.Parameter(
        uuid='c56650d0-252d-4b88-a645-fecd23dda1b6',
    )
    parameter_root.append_child(parameter_a)

    multiplexer_a = epcpm.canmodel.Multiplexer(
        name='TestMultiplexerA',
        identifier=1,
        uuid='bc872648-94d2-4bee-8a4f-5654cac93d38',
    )
    multiplexed_message.append_child(multiplexer_a)
    signal_a = epcpm.canmodel.Signal(
        name='SignalA',
        signed=True,
        parameter_uuid=parameter_a.uuid,
        uuid='cc5a8db2-b3da-4f2e-b8c6-83569b9c6824',
    )
    multiplexer_a.append_child(signal_a)

    expected += textwrap.dedent('''\
    Mux=TestMultiplexerA 0,8 1
    Var=CommonSignal signed 0,0 /d:0
    Var=SignalA signed 0,0 /d:0 /ln:"New Parameter"	// <rw:1:1>  <uuid:c56650d0-252d-4b88-a645-fecd23dda1b6>
    ''')

    assert tidy_sym(builder.gen()) == tidy_sym(expected)

    parameter_b = epyqlib.pm.parametermodel.Parameter(
        uuid='c6956fa5-15cc-48f4-a0a7-c4b691f14fec',
    )
    parameter_root.append_child(parameter_b)

    multiplexer_b = epcpm.canmodel.Multiplexer(
        name='TestMultiplexerB',
        identifier=2,
        uuid='7efd7e89-71f8-400b-b11d-7f49642309f8',
    )
    multiplexed_message.append_child(multiplexer_b)
    signal_b = epcpm.canmodel.Signal(
        name='SignalB',
        signed=True,
        parameter_uuid=parameter_b.uuid,
        uuid='b69c211b-0e0c-4ba1-947e-ea47a1baf11d',
    )
    multiplexer_b.append_child(signal_b)

    expected += textwrap.dedent('''\

    [TestMultiplexedMessage]
    DLC=0
    Mux=TestMultiplexerB 0,8 2
    Var=CommonSignal signed 0,0 /d:0
    Var=SignalB signed 0,0 /d:0 /ln:"New Parameter"	// <rw:1:1>  <uuid:c6956fa5-15cc-48f4-a0a7-c4b691f14fec>
    ''')

    assert tidy_sym(builder.gen()) == tidy_sym(expected)


def test_enumerations():
    project = epcpm.project.create_blank()
    parameter_root = project.models.parameters.root
    access_levels = epyqlib.pm.parametermodel.AccessLevels()
    access_level = epyqlib.pm.parametermodel.AccessLevel(value=0)
    access_levels.append_child(access_level)
    enumerations = epyqlib.pm.parametermodel.Enumerations()
    enumerations.append_child(access_levels)
    parameter_root.append_child(enumerations)
    can_root = project.models.can.root

    builder = epcpm.cantosym.builders.wrap(
        can_root,
        parameter_uuid_finder=project.models.can.node_from_uuid,
        parameter_model=project.models.parameters,
        access_levels=access_levels,
    )

    on_off = epyqlib.pm.parametermodel.Enumeration(name='OnOff')
    parameter_root.append_child(on_off)
    off = epyqlib.pm.parametermodel.Enumerator(name='off', value=0)
    on = epyqlib.pm.parametermodel.Enumerator(name='on', value=1)

    on_off.append_child(off)
    on_off.append_child(on)

    parameter = epyqlib.pm.parametermodel.Parameter(
        uuid='29cf3408-043e-4573-a511-0c7638688663',
    )
    parameter_root.append_child(parameter)

    message = epcpm.canmodel.Message()
    can_root.append_child(message)
    signal = epcpm.canmodel.Signal(
        enumeration_uuid=on_off.uuid,
        parameter_uuid=parameter.uuid,
    )
    message.append_child(signal)

    expected = textwrap.dedent('''\
    FormatVersion=5.0 // Do not edit this line!
    Title="canmatrix-Export"
    {ENUMS}
    enum OnOff(0="off", 1="on")
    
    {SENDRECEIVE}
    
    [NewMessage]
    ID=1FFFFFFFh
    Type=Extended
    DLC=0
    Var=NewSignal unsigned 0,0 /e:OnOff /d:0 /ln:"New Parameter"	// <rw:1:1>  <uuid:29cf3408-043e-4573-a511-0c7638688663>
    ''')

    assert tidy_sym(builder.gen()) == tidy_sym(expected)


def test_access_level():
    project = epcpm.project.create_blank()
    parameter_root = project.models.parameters.root
    can_root = project.models.can.root

    access_levels = epyqlib.pm.parametermodel.AccessLevels(
        name='AccessLevel',
        uuid='54bb11bb-a39d-4b30-a597-da3fbeda4d7a',
    )

    builder = epcpm.cantosym.builders.wrap(
        can_root,
        parameter_uuid_finder=project.models.can.node_from_uuid,
        parameter_model=project.models.parameters,
        access_levels=access_levels,
    )

    user = epyqlib.pm.parametermodel.Enumerator(
        name='user',
        value=0,
        uuid='b4c58cf9-3705-4e73-8e31-09f8da22ff61',
    )
    factory = epyqlib.pm.parametermodel.Enumerator(
        name='factory',
        value=1,
        uuid='f5dc405f-a0ed-4590-9553-bf3f1efec23a',
    )

    access_levels.append_child(user)
    access_levels.append_child(factory)

    enumerations = epyqlib.pm.parametermodel.Enumerations(
        name='Enumerations',
        uuid='1936c213-5230-454a-81f9-c8099046a19d',
    )
    parameter_root.append_child(enumerations)

    enumerations.append_child(access_levels)

    parameter = epyqlib.pm.parametermodel.Parameter(
        name='Factory Parameter',
        access_level_uuid=factory.uuid,
        uuid='40477147-cd0b-407a-9dc1-805d3205214b',
    )
    parameter_root.append_child(parameter)

    access_level_parameter = epyqlib.pm.parametermodel.Parameter(
        name='Access Parameter',
        uuid='908f42e7-7632-47bc-a8c7-eb1eca4e277c',
    )
    parameter_root.append_child(access_level_parameter)

    message = epcpm.canmodel.Message(

    )
    can_root.append_child(message)
    signal = epcpm.canmodel.Signal(
        name='FactorySignal',
        parameter_uuid=parameter.uuid,
        uuid='07047416-73e3-48bd-8c0c-85c1de078200',
    )
    message.append_child(signal)

    access_level_signal = epcpm.canmodel.Signal(
        name='AccessSignal',
        parameter_uuid=access_level_parameter.uuid,
        enumeration_uuid=access_levels.uuid,
        uuid='4f62704f-6548-4f80-b7b7-702f214a394b',
    )
    message.append_child(access_level_signal)

    expected = textwrap.dedent('''\
    FormatVersion=5.0 // Do not edit this line!
    Title="canmatrix-Export"
    {ENUMS}
    enum AccessLevel(0="user", 1="factory")

    {SENDRECEIVE}

    [NewMessage]
    ID=1FFFFFFFh
    Type=Extended
    DLC=0
    Var=FactorySignal unsigned 0,0 /d:0 /ln:"Factory Parameter"	// <rw:1:1> <factory>  <uuid:40477147-cd0b-407a-9dc1-805d3205214b>
    Var=AccessSignal unsigned 0,0 /e:AccessLevel /d:0 /ln:"Access Parameter"	// <rw:1:1>  <uuid:908f42e7-7632-47bc-a8c7-eb1eca4e277c>
    ''')

    assert tidy_sym(builder.gen()) == tidy_sym(expected)


def test_table():
    project = epcpm.project.loadp(here/'project'/'project.pmp')
    can_table, = project.models.can.root.nodes_by_attribute(
        attribute_value='First Table',
        attribute_name='name',
    )
    access_levels, = project.models.parameters.root.nodes_by_attribute(
        attribute_value='AccessLevel',
        attribute_name='name',
    )
    parameter_table = project.models.parameters.node_from_uuid(
        can_table.table_uuid,
    )
    parameter_table.update()

    can_table.update()

    builder = epcpm.cantosym.builders.wrap(
        project.models.can.root,
        parameter_uuid_finder=project.models.can.node_from_uuid,
        parameter_model=project.models.parameters,
        access_levels=access_levels,
    )

    result = builder.gen()
    print(result)

    expected = textwrap.dedent('''\
    FormatVersion=5.0 // Do not edit this line!
    Title="canmatrix-Export"
    {ENUMS}
    
    
    {SENDRECEIVE}
    
    [Tables]
    ID=1FFFFFFFh
    Type=Extended
    DLC=8
    Mux=First TableEO_0ET_0_ArrayOne 0,8 5 	// <table>
    Var=AO_0 unsigned 16,8 /d:0 /ln:"AO_0"	// <rw:1:1>  <uuid:715a25f9-39d3-4410-9f1f-119991ab2468>
    Var=AO_1 unsigned 24,8 /d:0 /ln:"AO_1"	// <rw:1:1>  <uuid:e1655e93-0f2a-45d8-b881-70e697e485f4>
    
    [Tables]
    DLC=8
    Mux=First TableEO_0ET_0_ArrayTwo_A 0,8 6 	// <table>
    Var=AT_0 unsigned 16,16 /d:0 /ln:"AT_0"	// <rw:1:1>  <uuid:a02b02ae-b430-490b-a8bf-a32d976fb35d>
    Var=AT_1 unsigned 32,16 /d:0 /ln:"AT_1"	// <rw:1:1>  <uuid:1999bee0-358a-475c-a122-c6171bfadf9a>
    Var=AT_2 unsigned 48,16 /d:0 /ln:"AT_2"	// <rw:1:1>  <uuid:7f41fcc0-d421-4dbd-a5d7-abac4cc3c868>
    
    [Tables]
    DLC=8
    Mux=First TableEO_0ET_0_ArrayTwo_B 0,8 7 	// <table>
    Var=AT_3 unsigned 16,16 /d:0 /ln:"AT_3"	// <rw:1:1>  <uuid:cc090c3e-ceac-4b6c-8b8d-d9a1fed5c80b>
    Var=AT_4 unsigned 32,16 /d:0 /ln:"AT_4"	// <rw:1:1>  <uuid:e887a919-2fe1-4801-a745-1b73b3032cfe>
    
    [Tables]
    DLC=8
    Mux=First TableEO_0ET_1_ArrayOne 0,8 8 	// <table>
    Var=AO_0 unsigned 16,8 /d:0 /ln:"AO_0"	// <rw:1:1>  <uuid:12889ce7-fa95-4707-bee1-e1db71330427>
    Var=AO_1 unsigned 24,8 /d:0 /ln:"AO_1"	// <rw:1:1>  <uuid:3366c5b9-0208-4c5e-818f-a72f0879f2cd>
    
    [Tables]
    DLC=8
    Mux=First TableEO_0ET_1_ArrayTwo_A 0,8 9 	// <table>
    Var=AT_0 unsigned 16,16 /d:0 /ln:"AT_0"	// <rw:1:1>  <uuid:9a89ce3e-be35-4905-8f93-fd29f952fde0>
    Var=AT_1 unsigned 32,16 /d:0 /ln:"AT_1"	// <rw:1:1>  <uuid:d2300447-d35d-4057-b37f-81da3a981378>
    Var=AT_2 unsigned 48,16 /d:0 /ln:"AT_2"	// <rw:1:1>  <uuid:68e82189-2b3b-429e-8bb7-36f6f800cce8>
    
    [Tables]
    DLC=8
    Mux=First TableEO_0ET_1_ArrayTwo_B 0,8 0Ah	// <table>
    Var=AT_3 unsigned 16,16 /d:0 /ln:"AT_3"	// <rw:1:1>  <uuid:7692c837-1e3f-4319-b4b7-4845cb6a9e3f>
    Var=AT_4 unsigned 32,16 /d:0 /ln:"AT_4"	// <rw:1:1>  <uuid:b8df3096-0051-4487-a819-d6c8c22328bc>
    
    [Tables]
    DLC=8
    Mux=First TableEO_0ET_2_ArrayOne 0,8 0Bh	// <table>
    Var=AO_0 unsigned 16,8 /d:0 /ln:"AO_0"	// <rw:1:1>  <uuid:7fc6837c-6f99-4980-9716-04fe088911a0>
    Var=AO_1 unsigned 24,8 /d:0 /ln:"AO_1"	// <rw:1:1>  <uuid:d53144c4-e5c7-47fd-8b82-caf617173dd5>
    
    [Tables]
    DLC=8
    Mux=First TableEO_0ET_2_ArrayTwo_A 0,8 0Ch	// <table>
    Var=AT_0 unsigned 16,16 /d:0 /ln:"AT_0"	// <rw:1:1>  <uuid:643978e1-b2fe-4f5a-bcf6-a55c7f4e8330>
    Var=AT_1 unsigned 32,16 /d:0 /ln:"AT_1"	// <rw:1:1>  <uuid:40b9f544-93e7-4cdc-ac37-251948cd8e99>
    Var=AT_2 unsigned 48,16 /d:0 /ln:"AT_2"	// <rw:1:1>  <uuid:29fd4dd1-1422-4f20-bf23-f15f4ee767f4>
    
    [Tables]
    DLC=8
    Mux=First TableEO_0ET_2_ArrayTwo_B 0,8 0Dh	// <table>
    Var=AT_3 unsigned 16,16 /d:0 /ln:"AT_3"	// <rw:1:1>  <uuid:65f14e81-d06d-4766-bdcb-aca8a47fd913>
    Var=AT_4 unsigned 32,16 /d:0 /ln:"AT_4"	// <rw:1:1>  <uuid:d5643fa0-2c4b-42a3-89c9-5a8a623a5fdf>
    
    [Tables]
    DLC=8
    Mux=First TableEO_0ET_3_ArrayOne 0,8 0Eh	// <table>
    Var=AO_0 unsigned 16,8 /d:0 /ln:"AO_0"	// <rw:1:1>  <uuid:6da0b3b7-00c9-4a71-a7bc-2eff01d9b072>
    Var=AO_1 unsigned 24,8 /d:0 /ln:"AO_1"	// <rw:1:1>  <uuid:5052fb2e-ff2b-45ed-8eb4-8fb20fa48f4b>
    
    [Tables]
    DLC=8
    Mux=First TableEO_0ET_3_ArrayTwo_A 0,8 0Fh	// <table>
    Var=AT_0 unsigned 16,16 /d:0 /ln:"AT_0"	// <rw:1:1>  <uuid:b5f3aa12-5374-448c-b459-d82b0423da62>
    Var=AT_1 unsigned 32,16 /d:0 /ln:"AT_1"	// <rw:1:1>  <uuid:34f19371-9da8-445e-9bdc-150609c55b0b>
    Var=AT_2 unsigned 48,16 /d:0 /ln:"AT_2"	// <rw:1:1>  <uuid:cdc54427-37cf-4369-bab4-c00fa3fffe1a>
    
    [Tables]
    DLC=8
    Mux=First TableEO_0ET_3_ArrayTwo_B 0,8 10h	// <table>
    Var=AT_3 unsigned 16,16 /d:0 /ln:"AT_3"	// <rw:1:1>  <uuid:463d5adc-6dcf-4821-8dad-efc86b611693>
    Var=AT_4 unsigned 32,16 /d:0 /ln:"AT_4"	// <rw:1:1>  <uuid:9ad7eec5-2c78-496d-ab4b-f57c1d1b8bb5>
    
    [Tables]
    DLC=8
    Mux=First TableEO_1ET_0_ArrayOne 0,8 11h	// <table>
    Var=AO_0 unsigned 16,8 /d:0 /ln:"AO_0"	// <rw:1:1>  <uuid:5b3ff5fa-8a91-4bcb-a39e-3fa3b299aabd>
    Var=AO_1 unsigned 24,8 /d:0 /ln:"AO_1"	// <rw:1:1>  <uuid:50601aeb-1838-4850-a6c1-285a819fc6c6>
    
    [Tables]
    DLC=8
    Mux=First TableEO_1ET_0_ArrayTwo_A 0,8 12h	// <table>
    Var=AT_0 unsigned 16,16 /d:0 /ln:"AT_0"	// <rw:1:1>  <uuid:c3764c5c-c14c-410a-9c85-981be157a7dc>
    Var=AT_1 unsigned 32,16 /d:0 /ln:"AT_1"	// <rw:1:1>  <uuid:24450cfa-7983-4786-9e89-e59c149b3221>
    Var=AT_2 unsigned 48,16 /d:0 /ln:"AT_2"	// <rw:1:1>  <uuid:14be748b-1768-4901-8214-d83b53e37747>
    
    [Tables]
    DLC=8
    Mux=First TableEO_1ET_0_ArrayTwo_B 0,8 13h	// <table>
    Var=AT_3 unsigned 16,16 /d:0 /ln:"AT_3"	// <rw:1:1>  <uuid:22900720-f67e-45fe-bd66-abb7872d1f5e>
    Var=AT_4 unsigned 32,16 /d:0 /ln:"AT_4"	// <rw:1:1>  <uuid:154ecbc4-0f8a-421a-b9fc-12859e2ff446>
    
    [Tables]
    DLC=8
    Mux=First TableEO_1ET_1_ArrayOne 0,8 14h	// <table>
    Var=AO_0 unsigned 16,8 /d:0 /ln:"AO_0"	// <rw:1:1>  <uuid:d2cac4d4-5c94-4ec9-8ec0-e006b42e934c>
    Var=AO_1 unsigned 24,8 /d:0 /ln:"AO_1"	// <rw:1:1>  <uuid:057d1ece-477e-4f42-a19f-beb640269d2b>
    
    [Tables]
    DLC=8
    Mux=First TableEO_1ET_1_ArrayTwo_A 0,8 15h	// <table>
    Var=AT_0 unsigned 16,16 /d:0 /ln:"AT_0"	// <rw:1:1>  <uuid:0345218d-4ae2-4652-b9ae-51a1a51db725>
    Var=AT_1 unsigned 32,16 /d:0 /ln:"AT_1"	// <rw:1:1>  <uuid:a676e7ae-69ba-4ab5-abfd-d7b6d1b61028>
    Var=AT_2 unsigned 48,16 /d:0 /ln:"AT_2"	// <rw:1:1>  <uuid:7a8dafb9-a89c-4159-bad6-cf08f3d47f6d>
    
    [Tables]
    DLC=8
    Mux=First TableEO_1ET_1_ArrayTwo_B 0,8 16h	// <table>
    Var=AT_3 unsigned 16,16 /d:0 /ln:"AT_3"	// <rw:1:1>  <uuid:016c8e13-6aa4-4ead-9fbc-66e95bef7166>
    Var=AT_4 unsigned 32,16 /d:0 /ln:"AT_4"	// <rw:1:1>  <uuid:b400a484-7edf-4519-aebb-cf6681c62ad6>
    
    [Tables]
    DLC=8
    Mux=First TableEO_1ET_2_ArrayOne 0,8 17h	// <table>
    Var=AO_0 unsigned 16,8 /d:0 /ln:"AO_0"	// <rw:1:1>  <uuid:bc12e308-f351-46f2-980b-3f69c7fe5f4d>
    Var=AO_1 unsigned 24,8 /d:0 /ln:"AO_1"	// <rw:1:1>  <uuid:bc2328ef-5ba9-46bc-8680-e32c9f231e43>
    
    [Tables]
    DLC=8
    Mux=First TableEO_1ET_2_ArrayTwo_A 0,8 18h	// <table>
    Var=AT_0 unsigned 16,16 /d:0 /ln:"AT_0"	// <rw:1:1>  <uuid:73ee2f19-2254-40f0-a8f8-96daf173803d>
    Var=AT_1 unsigned 32,16 /d:0 /ln:"AT_1"	// <rw:1:1>  <uuid:36eb85a9-d8f2-4e31-96f8-a17eb15442de>
    Var=AT_2 unsigned 48,16 /d:0 /ln:"AT_2"	// <rw:1:1>  <uuid:7e8f8bd2-e4fc-46c0-9982-d9fdaae35bbd>
    
    [Tables]
    DLC=8
    Mux=First TableEO_1ET_2_ArrayTwo_B 0,8 19h	// <table>
    Var=AT_3 unsigned 16,16 /d:0 /ln:"AT_3"	// <rw:1:1>  <uuid:1b3f4f7b-e3dd-4a6b-8f57-64cfd39ffbaf>
    Var=AT_4 unsigned 32,16 /d:0 /ln:"AT_4"	// <rw:1:1>  <uuid:bd327816-1be3-47f3-8d8c-348b2b50c89c>
    
    [Tables]
    DLC=8
    Mux=First TableEO_1ET_3_ArrayOne 0,8 1Ah	// <table>
    Var=AO_0 unsigned 16,8 /d:0 /ln:"AO_0"	// <rw:1:1>  <uuid:10a65f59-fe09-41ec-9c6b-4ccd3db653be>
    Var=AO_1 unsigned 24,8 /d:0 /ln:"AO_1"	// <rw:1:1>  <uuid:dc48c013-62e7-4665-bda6-a7d1a2164000>
    
    [Tables]
    DLC=8
    Mux=First TableEO_1ET_3_ArrayTwo_A 0,8 1Bh	// <table>
    Var=AT_0 unsigned 16,16 /d:0 /ln:"AT_0"	// <rw:1:1>  <uuid:45dca7f8-1f94-49bc-9028-83acb2ceb53d>
    Var=AT_1 unsigned 32,16 /d:0 /ln:"AT_1"	// <rw:1:1>  <uuid:fc375c2d-1b0a-446e-b467-27f278e71dd0>
    Var=AT_2 unsigned 48,16 /d:0 /ln:"AT_2"	// <rw:1:1>  <uuid:61e84991-52c1-4259-b2c4-0c904ca31f0a>
    
    [Tables]
    DLC=8
    Mux=First TableEO_1ET_3_ArrayTwo_B 0,8 1Ch	// <table>
    Var=AT_3 unsigned 16,16 /d:0 /ln:"AT_3"	// <rw:1:1>  <uuid:38edbaee-7580-41a8-82f0-9da8f494fb43>
    Var=AT_4 unsigned 32,16 /d:0 /ln:"AT_4"	// <rw:1:1>  <uuid:baed5be3-d32f-455b-aff4-4a353b28400b>
    ''')

    assert tidy_sym(result) == tidy_sym(expected)



import attr

import epcpm.c
import epcpm.parameterstointerface
import epcpm.sunspecmodel
import epcpm.sunspectoxlsx
import epyqlib.utils.general


builders = epyqlib.utils.general.TypeMap()


def export(c_path, h_path, sunspec_model, include_uuid_in_item):
    builder = builders.wrap(
        wrapped=sunspec_model.root,
        parameter_uuid_finder=sunspec_model.node_from_uuid,
        c_path=c_path,
        h_path=h_path,
        include_uuid_in_item=include_uuid_in_item,
    )

    c_path.parent.mkdir(parents=True, exist_ok=True)
    builder.gen()


def gen(self, first=0):
    lines = [[], []]

    for member in self.wrapped.children[first:]:
        try:
            builder = builders.wrap(
                wrapped=member,
                parameter_uuid_finder=self.parameter_uuid_finder,
                include_uuid_in_item=self.include_uuid_in_item,
            )
        except KeyError:
            continue
        new_lines = builder.gen()
        for ch_lines, ch_new_lines in zip(lines, new_lines):
            if len(ch_new_lines) > 0:
                ch_lines.extend(ch_new_lines)
                ch_lines.append('')

    return lines


@builders(epcpm.sunspecmodel.Root)
@attr.s
class Root:
    wrapped = attr.ib()
    parameter_uuid_finder = attr.ib()
    c_path = attr.ib()
    h_path = attr.ib()
    include_uuid_in_item = attr.ib()

    def gen(self):
        c_lines = [
            '#include "interface.h"',
            '#include "interfaceGen.h"',
            '#include "interfaceStructures_generated.h"',
            '#include "sunspecInterfaceGen.h"',
            '#include "sunspecInterfaceFunctions_generated.h"',
            '',
            '',
        ]

        include_guard = self.h_path.name.replace('.', '_').upper()

        h_lines = [
            f'#ifndef {include_guard}',
            f'#define {include_guard}',
            f'',
            f'#include "interface.h"',
            f'',
            f'',
        ]

        for member in self.wrapped.children:
            try:
                builder = builders.wrap(
                    wrapped=member,
                    parameter_uuid_finder=self.parameter_uuid_finder,
                    include_uuid_in_item=self.include_uuid_in_item,
                )
            except KeyError:
                continue
            more_c_lines, more_h_lines = builder.gen()
            c_lines.extend(more_c_lines)
            h_lines.extend(more_h_lines)

        h_lines.append('#endif')

        with self.c_path.open('w', newline='\n') as f:
            f.write(epcpm.c.format_nested_lists(c_lines).strip() + '\n')

        with self.h_path.open('w', newline='\n') as f:
            f.write(epcpm.c.format_nested_lists(h_lines).strip() + '\n')


@builders(epcpm.sunspecmodel.Model)
@attr.s
class SunSpecModel:
    wrapped = attr.ib()
    parameter_uuid_finder = attr.ib()
    include_uuid_in_item = attr.ib()

    def gen(self):
        return gen(self=self, first=1)


@builders(epcpm.sunspecmodel.HeaderBlock)
@builders(epcpm.sunspecmodel.FixedBlock)
@attr.s
class Block:
    wrapped = attr.ib()
    parameter_uuid_finder = attr.ib()
    include_uuid_in_item = attr.ib()

    gen = gen


@builders(epcpm.sunspecmodel.DataPointBitfield)
@attr.s
class DataPointBitfield:
    wrapped = attr.ib()
    parameter_uuid_finder = attr.ib()
    include_uuid_in_item = attr.ib()

    def gen(self):
        #TODO: CAMPid 07954360685417610543064316843160

        bits_per_modbus_register = 16
        bit_length = bits_per_modbus_register * self.wrapped.size

        name_uuid = str(self.wrapped.parameter_uuid).replace('-', '_')

        members = self.wrapped.children

        array_name = f'sunspecBitfieldItems_{name_uuid}'
        model = self.wrapped.tree_parent.tree_parent.id

        parameter = self.parameter_uuid_finder(self.wrapped.parameter_uuid)

        c_lines = []
        h_lines = []
        
        for member in members:
            member_name_uuid = str(member.parameter_uuid).replace('-', '_')
            variable_name = f'interfaceItem_variable_{member_name_uuid}'
            variable_type = epcpm.parameterstointerface.sunspec_types[
                self.parameter_uuid_finder(member.type_uuid).name
            ]

            h_lines.append(
                f'extern {variable_type} {variable_name};',
            )

            # TODO: initializer that doesn't assume int16_t
            c_lines.append(
                f'{variable_type} {variable_name} = 0;',
            )

        c_lines.append('')
        h_lines.append('')

        c_lines.extend([
            f'InterfaceItem_BitfieldMember {array_name}[{len(members)}] = {{',
            [
                [
                    f'[{i}] = {{',
                    [
                        f'.offset = {member.bit_offset},',
                        f'.length = {member.bit_length},',
                        f'.item = &interfaceItem_{str(member.parameter_uuid).replace("-", "_")},',
                    ],
                    f'}},',
                ]
                for i, member in enumerate(members)
            ],
            f'}};',
            f'',
            f'InterfaceItem_Bitfield interfaceItem_{name_uuid} = {{',
            [
                f'.common = {{',
                [   # TODO: generate a real complete common initializer
                    f'.sunspec = {{',
                    [
                        f'.getter = InterfaceItem_bitfield_{bit_length}_getter,',
                        f'.setter = InterfaceItem_bitfield_{bit_length}_setter,',
                        f'.variable = &sunspecInterface.model{model:05}.{parameter.abbreviation},',
                    ],
                    f'}},',
                ],
                f'}},',
                f'.members = {array_name},',
                f'.membersCount = {len(members)},',
            ],
            f'}};',
        ])

        h_lines.extend([
            f'extern InterfaceItem_BitfieldMember {array_name}[{len(members)}];',
            f'extern InterfaceItem_Bitfield interfaceItem_{name_uuid};',
        ])

        return c_lines, h_lines



# This file was automatically generated by sub\epyqlib\epyqlib\cli\generateversion.py
__version_build_time__ = '2020-06-03T17:07:09.659628'

__version__ = '2020.6'
__sha__ = '995791cd6f9fa07af75d4e30639b34ec0db74218'
__revision__ = '995791cd6f9fa07af75d4e30639b34ec0db74218'



import itertools
import math

import attr
import openpyxl

import epyqlib.pm.parametermodel
import epyqlib.utils.general

import epcpm.c
import epcpm.sunspecmodel


builders = epyqlib.utils.general.TypeMap()
enumeration_builders = epyqlib.utils.general.TypeMap()
enumerator_builders = epyqlib.utils.general.TypeMap()


data_point_fields = attr.fields(epcpm.sunspecmodel.DataPoint)
epc_enumerator_fields = attr.fields(
    epyqlib.pm.parametermodel.SunSpecEnumerator,
)
bitfield_fields = attr.fields(epcpm.sunspecmodel.DataPointBitfield)


def attr_fill(cls, value):
    return cls(**{field.name: value for field in attr.fields(cls) if field.init})


@attr.s
class Fields:
    field_type = attr.ib(default=None)
    applicable_point = attr.ib(default=None)
    address_offset = attr.ib(default=None)
    block_offset = attr.ib(default=None)
    size = attr.ib(default=None)
    name = attr.ib(default=None)
    label = attr.ib(default=None)
    value = attr.ib(default=None)
    type = attr.ib(default=None)
    units = attr.ib(default=None)
    scale_factor = attr.ib(default=None)
    read_write = attr.ib(default=None)
    mandatory = attr.ib(default=None)
    description = attr.ib(default=None)
    notes = attr.ib(default=None)
    modbus_address = attr.ib(default=None)
    get = attr.ib(default=None)
    set = attr.ib(default=None)
    item = attr.ib(default=None)

    def as_filtered_tuple(self, filter_):
        return tuple(
            value
            for value, f in zip(attr.astuple(self), attr.astuple(filter_))
            if f
        )


field_names = Fields(
    field_type='Field Type',
    applicable_point='Applicable Point',
    address_offset='Address Offset',
    block_offset='Block Offset',
    size='Size',
    name='Name',
    label='Label',
    value='Value',
    type='Type',
    units='Units',
    scale_factor='SF',
    read_write='R/W',
    mandatory='Mandatory M/O',
    description='Description',
    notes='Notes',
    modbus_address='Modbus Address',
    get='get',
    set='set',
    item='item',
)


point_fields = Fields(
    block_offset=data_point_fields.block_offset,
    size=data_point_fields.size,
    type=data_point_fields.type_uuid,
    scale_factor=data_point_fields.factor_uuid,
    units=data_point_fields.units,
)

bitfield_fields = Fields(
    block_offset=bitfield_fields.block_offset,
    size=bitfield_fields.size,
    type=bitfield_fields.type_uuid,
)

enumerator_fields = Fields(
    name=epc_enumerator_fields.abbreviation,
    label=epc_enumerator_fields.label,
    description=epc_enumerator_fields.description,
    notes=epc_enumerator_fields.notes,
    value=epc_enumerator_fields.value,
    field_type=epc_enumerator_fields.type,
)


def export(
        path,
        sunspec_model,
        parameters_model,
        column_filter=None,
        skip_sunspec=False
    ):
    if column_filter is None:
        column_filter = attr_fill(Fields, True)

    builder = epcpm.sunspectoxlsx.builders.wrap(
        wrapped=sunspec_model.root,
        parameter_uuid_finder=sunspec_model.node_from_uuid,
        parameter_model=parameters_model,
        skip_sunspec=skip_sunspec,
        column_filter=column_filter,
    )

    workbook = builder.gen()

    path.parent.mkdir(parents=True, exist_ok=True)
    workbook.save(path)


@builders(epcpm.sunspecmodel.Root)
@attr.s
class Root:
    wrapped = attr.ib()
    column_filter = attr.ib()
    skip_sunspec = attr.ib(default=False)
    parameter_uuid_finder = attr.ib(default=None)
    parameter_model = attr.ib(default=None)

    def gen(self):
        workbook = openpyxl.Workbook()
        workbook.remove(workbook.active)

        workbook.create_sheet('License Agreement')
        workbook.create_sheet('Summary')
        workbook.create_sheet('Index')

        if not self.skip_sunspec:
            children = sorted(
                self.wrapped.children,
                key=lambda child: (
                    0 if isinstance(child, epcpm.sunspecmodel.Model) else 1,
                    getattr(child, 'id', math.inf),
                )
            )

            model_offset = 2 #account for starting 'SunS'
            for model in children:
                if isinstance(model, epcpm.sunspecmodel.Table):
                    # TODO: for now, implement it soon...
                    continue

                worksheet = workbook.create_sheet()

                model_offset += builders.wrap(
                    wrapped=model,
                    worksheet=worksheet,
                    padding_type=self.parameter_model.list_selection_roots['sunspec types'].child_by_name('pad'),
                    parameter_uuid_finder=self.parameter_uuid_finder,
                    column_filter=self.column_filter,
                    model_offset=model_offset,
                ).gen()

        return workbook


@builders(epcpm.sunspecmodel.Model)
@attr.s
class Model:
    wrapped = attr.ib()
    worksheet = attr.ib()
    padding_type = attr.ib()
    column_filter = attr.ib()
    model_offset = attr.ib() #starting Modbus address for the model
    parameter_uuid_finder = attr.ib(default=None)

    def gen(self):
        self.worksheet.title = str(self.wrapped.id)
        self.worksheet.append(field_names.as_filtered_tuple(self.column_filter))

        self.wrapped.children[0].check_offsets_and_length()

        overall_length = sum(
            child.check_offsets_and_length()
            for child in self.wrapped.children[1:]
        )

        add_padding = (overall_length % 2) == 1
        if add_padding:
            overall_length += 1

        accumulated_length = 0

        rows = []

        model_types = ['Header', 'Fixed Block', 'Repeating Block']

        zipped = zip(enumerate(self.wrapped.children), model_types)
        for (i, child), model_type in zipped:
            builder = builders.wrap(
                wrapped=child,
                add_padding=add_padding and i == 1,
                padding_type=self.padding_type,
                model_type=model_type,
                model_id=self.wrapped.id,
                model_offset=self.model_offset,
                parameter_uuid_finder=self.parameter_uuid_finder,
                address_offset=accumulated_length,
            )

            built_rows, block_length = builder.gen()
            accumulated_length += block_length
            if len(built_rows) > 0:
                rows.extend(built_rows)
                rows.append(Fields())

        for i, row in enumerate(rows):
            if i == 0:
                row.value = self.wrapped.id
            elif i == 1:
                row.value = overall_length

            self.worksheet.append(row.as_filtered_tuple(self.column_filter))

        for block in self.wrapped.children:
            builder = enumeration_builders.wrap(
                wrapped=block,
                parameter_uuid_finder=self.parameter_uuid_finder,
            )
            rows = builder.gen()

            for row in rows:
                self.worksheet.append(
                    row.as_filtered_tuple(self.column_filter),
                )

        return overall_length + 2 #add header length


@builders(epcpm.sunspecmodel.Table)
@attr.s
class Table:
    wrapped = attr.ib()
    worksheet = attr.ib()
    padding_type = attr.ib()
    column_filter = attr.ib()
    parameter_uuid_finder = attr.ib(default=None)

    def gen(self):
        return []


@enumeration_builders(epcpm.sunspecmodel.DataPoint)
@attr.s
class Enumeration:
    wrapped = attr.ib()
    point = attr.ib()
    parameter_uuid_finder = attr.ib(default=None)

    def gen(self):
        rows = []

        enumeration_uuid = getattr(self.wrapped, 'enumeration_uuid', None)
        if enumeration_uuid is None:
            return rows

        enumeration = self.parameter_uuid_finder(enumeration_uuid)

        for enumerator in enumeration.children:
            builder = enumerator_builders.wrap(
                wrapped=enumerator,
                point=self.point,
                parameter_uuid_finder=self.parameter_uuid_finder,
            )
            rows.append(builder.gen())

        return rows


@enumerator_builders(epyqlib.pm.parametermodel.SunSpecEnumerator)
@attr.s
class Enumerator:
    wrapped = attr.ib()
    point = attr.ib()
    parameter_uuid_finder = attr.ib(default=None)

    # TODO: CAMPid 07397546759269756456100183066795496952476951653
    def gen(self):
        row = Fields()

        for name, field in attr.asdict(enumerator_fields).items():
            if field is None:
                continue

            setattr(row, name, getattr(self.wrapped, field.name))

        if row.name is None:
            row.name = self.wrapped.name

        parameter = self.parameter_uuid_finder(self.point.parameter_uuid)
        row.applicable_point = parameter.abbreviation

        field_type_parameter = self.parameter_uuid_finder(self.point.type_uuid)
        row.field_type = field_type_parameter.name

        return row


def build_uuid_scale_factor_dict(points, parameter_uuid_finder):
    scale_factor_from_uuid = {}
    for point in points:
        if point.type_uuid is None:
            continue

        type_node = parameter_uuid_finder(point.type_uuid)

        if type_node is None:
            continue

        if type_node.name == 'sunssf':
            scale_factor_from_uuid[point.uuid] = point

    return scale_factor_from_uuid


@builders(epcpm.sunspecmodel.TableRepeatingBlock)
@builders(epcpm.sunspecmodel.HeaderBlock)
@builders(epcpm.sunspecmodel.FixedBlock)
@attr.s
class Block:
    wrapped = attr.ib()
    add_padding = attr.ib()
    padding_type = attr.ib()
    model_type = attr.ib()
    model_id = attr.ib()
    model_offset = attr.ib()
    address_offset = attr.ib()
    repeating_block_reference = attr.ib(default=None)
    parameter_uuid_finder = attr.ib(default=None)
    is_table = attr.ib(default=False)

    def gen(self):
        # TODO: CAMPid 07548795421667967542697543743987

        scale_factor_from_uuid = build_uuid_scale_factor_dict(
            points=self.wrapped.children,
            parameter_uuid_finder=self.parameter_uuid_finder,
        )

        rows = []

        points = list(self.wrapped.children)

        if self.add_padding:
            point = epcpm.sunspecmodel.DataPoint(
                type_uuid=self.padding_type.uuid,
                block_offset=(
                    self.wrapped.children[-1].block_offset
                    + self.wrapped.children[-1].size
                ),
                size=self.padding_type.value,
            )
            # TODO: ack!  just to get the address offset calculated but
            #       not calling append_child() because i don't want to shove
            #       this into the model.  :[
            point.tree_parent = self.wrapped
            points.append(point)

        summed_increments = 0

        for child in points:
            builder = builders.wrap(
                wrapped=child,
                model_type=self.model_type,
                scale_factor_from_uuid=scale_factor_from_uuid,
                parameter_uuid_finder=self.parameter_uuid_finder,
                model_id=self.model_id,
                model_offset=self.model_offset,
                is_table=self.is_table,
                repeating_block_reference=self.repeating_block_reference,
                address_offset=self.address_offset + summed_increments,
            )
            built_rows, address_offset_increment = builder.gen()
            summed_increments += address_offset_increment
            rows.append(built_rows)

        return rows, summed_increments


@builders(epcpm.sunspecmodel.DataPointBitfield)
@attr.s
class DataPointBitfield:
    wrapped = attr.ib()
    model_type = attr.ib()
    scale_factor_from_uuid = attr.ib()
    parameter_uuid_finder = attr.ib()
    model_id = attr.ib()
    model_offset = attr.ib()
    is_table = attr.ib()
    repeating_block_reference = attr.ib()
    address_offset = attr.ib()

    def gen(self):
        row = Fields()
        row.address_offset = self.address_offset
        row.modbus_address = self.model_offset + self.address_offset
        row.field_type = self.model_type

        for name, field in attr.asdict(bitfield_fields).items():
            if field is None:
                continue

            setattr(row, name, getattr(self.wrapped, field.name))

        if row.type is not None:
            row.type = self.parameter_uuid_finder(row.type).name

        if self.wrapped.parameter_uuid is not None:
            parameter = self.parameter_uuid_finder(self.wrapped.parameter_uuid)

            row.label = parameter.name
            row.name = parameter.abbreviation
            row.notes = '' if parameter.notes is None else parameter.notes
            row.notes = f'{row.notes}  <uuid:{parameter.uuid}>'.strip()

            if row.units is None:
                row.units = parameter.units
            row.description = parameter.comment
            row.read_write = 'R' if parameter.read_only else 'RW'

        uses_interface_item = (
                isinstance(parameter, epyqlib.pm.parametermodel.Parameter)
                and parameter.uses_interface_item()
        )

        getter = []
        setter = []

        # TODO: should we just require that it does and assume etc?
        if uses_interface_item:
            # TODO: CAMPid 9685439641536675431653179671436
            parameter_uuid = str(parameter.uuid).replace('-', '_')
            item_name = f'interfaceItem_{parameter_uuid}'

            getter.extend([
                f'{item_name}.common.sunspec.getter(',
                [
                    f'(InterfaceItem_void *) &{item_name},',
                    f'Meta_Value',
                ],
                f');',
            ])
            setter.extend([
                f'{item_name}.common.sunspec.setter(',
                [
                    f'(InterfaceItem_void *) &{item_name},',
                    f'true,',
                    f'Meta_Value',
                ],
                f');',
            ])

        if len(getter) > 0:
            # TODO: what if write-only?
            row.get = epcpm.c.format_nested_lists(getter)

        if not parameter.read_only:
            row.set = epcpm.c.format_nested_lists(setter)
        else:
            row.set = None

        return row, row.size


@enumeration_builders(epcpm.sunspecmodel.DataPointBitfield)
@attr.s
class DataPointBitfield:
    wrapped = attr.ib()
    point = attr.ib()
    parameter_uuid_finder = attr.ib(default=None)

    def gen(self):
        rows = []

        for enumerator in self.wrapped.children:
            builder = enumerator_builders.wrap(
                wrapped=enumerator,
                point=self.point,
                parameter_uuid_finder=self.parameter_uuid_finder,
            )
            rows.append(builder.gen())

        return rows


@enumerator_builders(epcpm.sunspecmodel.DataPointBitfieldMember)
@attr.s
class DataPointBitfieldMember:
    wrapped = attr.ib()
    point = attr.ib()
    parameter_uuid_finder = attr.ib(default=None)

    # TODO: CAMPid 07397546759269756456100183066795496952476951653
    def gen(self):
        field_parameter = self.parameter_uuid_finder(self.point.parameter_uuid)
        member_parameter = self.parameter_uuid_finder(self.wrapped.parameter_uuid)

        length = self.point.size * 16

        row = Fields(
            field_type=f'bitfield{length}',
            value=self.wrapped.bit_offset,
            applicable_point=field_parameter.abbreviation,
            name=member_parameter.abbreviation,
            label=member_parameter.name,
            description=member_parameter.comment,
            notes=member_parameter.notes,
        )

        return row


@enumeration_builders(epcpm.sunspecmodel.HeaderBlock)
@enumeration_builders(epcpm.sunspecmodel.FixedBlock)
@enumeration_builders(epcpm.sunspecmodel.TableRepeatingBlockReference)
@attr.s
class GenericEnumeratorBuilder:
    wrapped = attr.ib()
    parameter_uuid_finder = attr.ib(default=None)

    def gen(self):
        rows = []

        for child in self.wrapped.children:
            builder = enumeration_builders.wrap(
                wrapped=child,
                point=child,
                parameter_uuid_finder=self.parameter_uuid_finder,
            )

            new_rows = builder.gen()

            if len(new_rows) > 0:
                rows.extend(new_rows)
                rows.append(Fields())

        return rows


@enumeration_builders(
    epcpm.sunspecmodel.TableRepeatingBlockReferenceDataPointReference,
)
@attr.s
class TableRepeatingBlockReferenceDataPointReferenceEnumerationBuilder:
    wrapped = attr.ib()
    point = attr.ib()
    parameter_uuid_finder = attr.ib(default=None)

    def gen(self):
        builder = enumeration_builders.wrap(
            wrapped=self.wrapped.original,
            point=self.point.original,
            parameter_uuid_finder=self.parameter_uuid_finder,
        )

        return builder.gen()


@builders(epcpm.sunspecmodel.TableRepeatingBlockReference)
@attr.s
class TableRepeatingBlockReference:
    wrapped = attr.ib()
    add_padding = attr.ib()
    padding_type = attr.ib()
    model_type = attr.ib()
    model_id = attr.ib()
    model_offset = attr.ib()
    address_offset = attr.ib()
    parameter_uuid_finder = attr.ib(default=None)

    def gen(self):
        builder = builders.wrap(
            wrapped=self.wrapped.original,
            model_type=self.model_type,
            parameter_uuid_finder=self.parameter_uuid_finder,
            add_padding=self.add_padding,
            padding_type=self.padding_type,
            model_id=self.model_id,
            model_offset=self.model_offset,
            is_table=True,
            repeating_block_reference=self.wrapped,
            address_offset=self.address_offset,
        )

        return builder.gen()


@builders(epcpm.sunspecmodel.DataPoint)
@attr.s
class Point:
    wrapped = attr.ib()
    scale_factor_from_uuid = attr.ib()
    model_type = attr.ib()
    model_id = attr.ib()
    model_offset = attr.ib()
    is_table = attr.ib()
    address_offset = attr.ib()
    repeating_block_reference = attr.ib()
    parameter_uuid_finder = attr.ib(default=None)

    # TODO: CAMPid 07397546759269756456100183066795496952476951653
    def gen(self):
        row = Fields()
        row.address_offset = self.address_offset
        row.modbus_address = self.model_offset + self.address_offset

        for name, field in attr.asdict(point_fields).items():
            if field is None:
                continue

            setattr(row, name, getattr(self.wrapped, field.name))

        if self.repeating_block_reference is not None:
            target = (
                self.parameter_uuid_finder(self.wrapped.parameter_uuid)
                .original
            )
            is_array_element = isinstance(
                target,
                epyqlib.pm.parametermodel.ArrayParameterElement,
            )
            if is_array_element:
                target = target.tree_parent.children[0]

            references = [
                child
                for child in self.repeating_block_reference.children
                if target.uuid == child.parameter_uuid
            ]
            if len(references) > 0:
                reference, = references

                if reference.factor_uuid is not None:
                    row.scale_factor = self.parameter_uuid_finder(self.parameter_uuid_finder(reference.factor_uuid).parameter_uuid).abbreviation
        else:
            if row.scale_factor is not None:
                row.scale_factor = (
                    self.parameter_uuid_finder(
                        self.scale_factor_from_uuid[row.scale_factor].parameter_uuid).abbreviation
                )

        if row.type is not None:
            row.type = self.parameter_uuid_finder(row.type).name

        row.mandatory = 'M' if self.wrapped.mandatory else 'O'

        if self.wrapped.parameter_uuid is not None:
            parameter = self.parameter_uuid_finder(self.wrapped.parameter_uuid)

            row.label = parameter.name
            row.name = parameter.abbreviation
            row.notes = '' if parameter.notes is None else parameter.notes
            to_tree = list(itertools.takewhile(
                lambda node: not isinstance(node, epyqlib.pm.parametermodel.Table),
                parameter.ancestors(),
            ))
            tree = to_tree[-1]
            if isinstance(tree, epyqlib.pm.parametermodel.TableGroupElement):
                # naturally sorted by traversal...  hopefully
                all_of_them = tree.nodes_by_filter(
                    filter=lambda node: node.original == parameter.original,
                    collection=[],
                )
                uuids_string = ' '.join(
                    (f'<uuid:{parameter.uuid}>' for parameter in all_of_them),
                )
                row.notes = f'{row.notes}  {uuids_string}'.strip()
            else:
                row.notes = f'{row.notes}  <uuid:{parameter.uuid}>'.strip()

            if row.units is None:
                row.units = parameter.units
            row.description = parameter.comment
            row.read_write = 'R' if parameter.read_only else 'RW'

            meta = '[Meta_Value]'

            getter = []
            setter = []

            uses_interface_item = (
                isinstance(parameter, epyqlib.pm.parametermodel.Parameter)
                and parameter.uses_interface_item()
            )

            hand_coded_getter_function_name = getter_name(
                parameter=parameter,
                model_id=self.model_id,
                is_table=self.is_table,
            )

            hand_coded_setter_function_name = setter_name(
                parameter=parameter,
                model_id=self.model_id,
                is_table=self.is_table,
            )

            if not uses_interface_item and not self.wrapped.not_implemented:
                if row.scale_factor is not None:
                    scale_factor_updater_name = (
                        f'getSUNSPEC_MODEL{self.model_id}_{row.scale_factor}'
                    )

                    f = f'{scale_factor_updater_name}();'
                    get_scale_factor = f.format(
                        model_id=self.model_id,
                        abbreviation=row.scale_factor,
                    )
                    getter.append(get_scale_factor)
                    setter.append(get_scale_factor)

                getter.append(f'{hand_coded_getter_function_name}();')

            sunspec_model_variable = f'sunspecInterface.model{self.model_id}'

            sunspec_variable = (
                f'{sunspec_model_variable}.{parameter.abbreviation}'
            )

            if row.type == 'pad':
                getter.append(
                    f'{sunspec_variable} = 0x8000;'
                )
            elif self.wrapped.not_implemented:
                value = {
                    'int16': 'INT16_C(0x8000)',
                    'uint16': 'UINT16_C(0xffff)',
                    'acc16': 'UINT16_C(0x0000)',
                    'enum16': 'UINT16_C(0xffff)',
                    'bitfield16': 'UINT16_C(0xffff)',
                    'int32': 'sunspecInt32ToSS32_returns(INT32_C(0x80000000))',
                    'uint32': 'sunspecUint32ToSSU32_returns(UINT32_C(0xffffffff))',
                    'acc32': 'sunspecUint32ToSSU32_returns(UINT32_C(0x00000000))',
                    'enum32': 'sunspecUint32ToSSU32_returns(UINT32_C(0xffffffff))',
                    'bitfield32': 'sunspecUint32ToSSU32_returns(UINT32_C(0xffffffff))',
                    'ipaddr': 'sunspecUint32ToSSU32_returns(UINT32_C(0x00000000))',
                    'int64': 'sunspecInt64ToSS64_returns(INT64_C(0x8000000000000000))',
                    # yes, acc64 seems to be an int64, not a uint64
                    'acc64': 'sunspecInt64ToSS64_returns(INT64_C(0x0000000000000000))',
                    # 'ipv6addr': 'INT128_C(0x00000000000000000000000000000000)',
                    # 'float32': 'NAN',
                    'sunssf': 'INT16_C(0x8000)',
                    'string': 'UINT16_C(0x0000)',
                }[row.type]
                if row.type == 'string':
                    getter.extend([
                        f'for (size_t i = 0; i < LENGTHOF({sunspec_variable}); i++) {{',
                        [
                            f'{sunspec_variable}[i] = {value};'
                        ],
                        '}',
                    ])
                elif row.type.startswith('bitfield'):
                    getter.append(
                        f'{sunspec_variable}.raw = {value};'
                    )
                    # # below because parsesunspec only detects bitfields
                    # # if they have values
                    # if self.wrapped.enumeration_uuid is not None:
                    #     getter.append(
                    #         f'{sunspec_variable}.raw = {value};'
                    #     )
                    # else:
                    #     getter.append(
                    #         f'*((uint{row.type[-2:]}_t*) &{sunspec_variable})'
                    #         f' = {value};'
                    #     )
                else:
                    getter.append(
                        f'{sunspec_variable} = {value};'
                    )

                setter.append('// point not implemented, do nothing')
            elif parameter.nv_format is not None:
                internal_variable = parameter.nv_format.format(meta)

                # TODO: CAMPid 075780541068182645821856068542023499
                converter = {
                    'uint32': {
                        'get': 'sunspecUint32ToSSU32',
                        'set': 'sunspecSSU32ToUint32',
                    },
                    'int32': {
                        # TODO: add this to embedded?
                        # 'get': 'sunspecInt32ToSSS32',
                        'set': 'sunspecSSS32ToInt32',
                    },
                }.get(row.type)

                if converter is not None:
                    get_converter = converter['get']
                    set_converter = converter['set']

                    get_cast = ''
                    set_cast = ''
                    if parameter.nv_cast:
                        set_cast = f'(__typeof__({internal_variable})) '
                        get_type = {
                            'uint32': 'uint32_t',
                        }[row.type]
                        get_cast = f'({get_type})'

                    getter.extend([
                        f'{get_converter}(',
                        [
                            f'&{sunspec_variable},',
                            f'{get_cast}{internal_variable}',
                        ],
                        ');',
                    ])
                    setter.extend([
                        f'{internal_variable} = {set_cast}{set_converter}(',
                        [
                            f'&{sunspec_variable}',
                        ],
                        ');',
                    ])
                else:
                    getter.append(adjust_assignment(
                        left_hand_side=sunspec_variable,
                        right_hand_side=internal_variable,
                        sunspec_model_variable=sunspec_model_variable,
                        scale_factor=row.scale_factor,
                        internal_scale=parameter.internal_scale_factor,
                        parameter=parameter,
                        factor_operator='*',
                    ))

                    setter.append(adjust_assignment(
                        left_hand_side=internal_variable,
                        right_hand_side=sunspec_variable,
                        sunspec_model_variable=sunspec_model_variable,
                        scale_factor=row.scale_factor,
                        internal_scale=parameter.internal_scale_factor,
                        parameter=parameter,
                        factor_operator='/',
                    ))

                # minimum_variable = parameter.nv_format.format('[Meta_Min]')
                # maximum_variable = parameter.nv_format.format('[Meta_Max]')
            elif uses_interface_item:
                # TODO: CAMPid 9685439641536675431653179671436
                parameter_uuid = str(parameter.uuid).replace('-', '_')
                item_name = f'interfaceItem_{parameter_uuid}'

                getter.extend([
                    f'{item_name}.common.sunspec.getter(',
                    [
                        f'(InterfaceItem_void *) &{item_name},',
                        f'Meta_Value',
                    ],
                    f');',
                ])
                setter.extend([
                    f'{item_name}.common.sunspec.setter(',
                    [
                        f'(InterfaceItem_void *) &{item_name},',
                        f'true,',
                        f'Meta_Value',
                    ],
                    f');',
                ])
            else:
                if getattr(parameter, 'sunspec_getter', None) is not None:
                    getter.append(
                        parameter.sunspec_getter.format(
                            interface=sunspec_variable,
                        )
                    )

                if getattr(parameter, 'sunspec_setter', None) is not None:
                    setter.append(
                        parameter.sunspec_setter.format(
                            interface=sunspec_variable,
                        )
                    )

            row.get = epcpm.c.format_nested_lists(getter)

            if not uses_interface_item and not self.wrapped.not_implemented:
                setter.append(f'{hand_coded_setter_function_name}();')

            if not parameter.read_only:
                row.set = epcpm.c.format_nested_lists(setter)
            else:
                row.set = None

        row.field_type = self.model_type

        if self.parameter_uuid_finder(self.wrapped.type_uuid).name == 'pad':
            row.name = 'Pad'
            row.description = 'Force even alignment'
            row.read_write = 'R'
            row.mandatory = 'O'

        return row, row.size


def adjust_assignment(
        left_hand_side,
        right_hand_side,
        sunspec_model_variable,
        scale_factor,
        internal_scale,
        parameter,
        factor_operator,
):
    if scale_factor is not None:
        scale_factor_variable = f'{sunspec_model_variable}.{scale_factor}'
        # TODO: what about positive scalings?
        # factor = f'(P99_IPOW(-{scale_factor_variable}, 10))'
        # TODO: we really don't want doubles here, do we?
        # factor = f'(pow(10, -{scale_factor_variable}))'

        opposite = '' if factor_operator == '*' else '-'

        right_hand_side = (
            f'(sunspecScale({right_hand_side},'
            f' {opposite}({scale_factor_variable} + {internal_scale})))'
        )

    if parameter.nv_cast:
        right_hand_side = f'((__typeof__({left_hand_side})) {right_hand_side})'

    result = f'{left_hand_side} = {right_hand_side};'

    return result


def getter_setter_name(get_set, parameter, model_id, is_table):
    if is_table:
        table_option = '_{table_option}'
    else:
        table_option = ''

    format_string = (
        '{get_set}SunspecModel{model_id}{table_option}_{abbreviation}'
    )

    return format_string.format(
        get_set=get_set,
        model_id=model_id,
        abbreviation=parameter.abbreviation,
        table_option=table_option,
    )


def getter_name(parameter, model_id, is_table):
    return getter_setter_name(
        get_set='get',
        parameter=parameter,
        model_id=model_id,
        is_table=is_table,
    )


def setter_name(parameter, model_id, is_table):
    return getter_setter_name(
        get_set='set',
        parameter=parameter,
        model_id=model_id,
        is_table=is_table,
    )



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'mainwindow.ui'
#
# Created by: PyQt5 UI code generator 5.13.0
#
# WARNING! All changes made in this file will be lost!


from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_MainWindow(object):
    def setupUi(self, MainWindow):
        MainWindow.setObjectName("MainWindow")
        MainWindow.resize(1200, 900)
        self.centralwidget = QtWidgets.QWidget(MainWindow)
        self.centralwidget.setObjectName("centralwidget")
        self.gridLayout_5 = QtWidgets.QGridLayout(self.centralwidget)
        self.gridLayout_5.setObjectName("gridLayout_5")
        self.splitter = QtWidgets.QSplitter(self.centralwidget)
        self.splitter.setOrientation(QtCore.Qt.Horizontal)
        self.splitter.setObjectName("splitter")
        self.gridLayoutWidget = QtWidgets.QWidget(self.splitter)
        self.gridLayoutWidget.setObjectName("gridLayoutWidget")
        self.gridLayout = QtWidgets.QGridLayout(self.gridLayoutWidget)
        self.gridLayout.setContentsMargins(0, 0, 0, 0)
        self.gridLayout.setObjectName("gridLayout")
        self.parameter_view = QtWidgets.QTreeView(self.gridLayoutWidget)
        self.parameter_view.setAlternatingRowColors(True)
        self.parameter_view.setObjectName("parameter_view")
        self.gridLayout.addWidget(self.parameter_view, 2, 0, 1, 1)
        self.parameter_search_box = SearchBox(self.gridLayoutWidget)
        self.parameter_search_box.setObjectName("parameter_search_box")
        self.gridLayout.addWidget(self.parameter_search_box, 0, 0, 1, 1)
        self.tabWidget = QtWidgets.QTabWidget(self.splitter)
        self.tabWidget.setObjectName("tabWidget")
        self.tab_2 = QtWidgets.QWidget()
        self.tab_2.setObjectName("tab_2")
        self.horizontalLayout_2 = QtWidgets.QHBoxLayout(self.tab_2)
        self.horizontalLayout_2.setObjectName("horizontalLayout_2")
        self.gridLayout_2 = QtWidgets.QGridLayout()
        self.gridLayout_2.setObjectName("gridLayout_2")
        self.can_view = QtWidgets.QTreeView(self.tab_2)
        self.can_view.setAlternatingRowColors(True)
        self.can_view.setObjectName("can_view")
        self.gridLayout_2.addWidget(self.can_view, 1, 0, 1, 1)
        self.can_search_box = SearchBox(self.tab_2)
        self.can_search_box.setObjectName("can_search_box")
        self.gridLayout_2.addWidget(self.can_search_box, 0, 0, 1, 1)
        self.horizontalLayout_2.addLayout(self.gridLayout_2)
        self.tabWidget.addTab(self.tab_2, "")
        self.tab_21 = QtWidgets.QWidget()
        self.tab_21.setObjectName("tab_21")
        self.horizontalLayout_21 = QtWidgets.QHBoxLayout(self.tab_21)
        self.horizontalLayout_21.setObjectName("horizontalLayout_21")
        self.gridLayout_3 = QtWidgets.QGridLayout()
        self.gridLayout_3.setObjectName("gridLayout_3")
        self.sunspec_view = QtWidgets.QTreeView(self.tab_21)
        self.sunspec_view.setObjectName("sunspec_view")
        self.gridLayout_3.addWidget(self.sunspec_view, 1, 0, 1, 1)
        self.sunspec_search_box = SearchBox(self.tab_21)
        self.sunspec_search_box.setObjectName("sunspec_search_box")
        self.gridLayout_3.addWidget(self.sunspec_search_box, 0, 0, 1, 1)
        self.horizontalLayout_21.addLayout(self.gridLayout_3)
        self.tabWidget.addTab(self.tab_21, "")
        self.tab = QtWidgets.QWidget()
        self.tab.setObjectName("tab")
        self.horizontalLayout = QtWidgets.QHBoxLayout(self.tab)
        self.horizontalLayout.setObjectName("horizontalLayout")
        self.gridLayout_4 = QtWidgets.QGridLayout()
        self.gridLayout_4.setObjectName("gridLayout_4")
        self.value_set_view = QtWidgets.QTreeView(self.tab)
        self.value_set_view.setObjectName("value_set_view")
        self.gridLayout_4.addWidget(self.value_set_view, 1, 0, 1, 1)
        self.value_set_search_box = SearchBox(self.tab)
        self.value_set_search_box.setObjectName("value_set_search_box")
        self.gridLayout_4.addWidget(self.value_set_search_box, 0, 0, 1, 1)
        self.horizontalLayout.addLayout(self.gridLayout_4)
        self.tabWidget.addTab(self.tab, "")
        self.gridLayoutWidget22 = QtWidgets.QWidget(self.splitter)
        self.gridLayoutWidget22.setObjectName("gridLayoutWidget22")
        self.gridLayout22 = QtWidgets.QGridLayout(self.gridLayoutWidget22)
        self.gridLayout22.setContentsMargins(0, 0, 0, 0)
        self.gridLayout22.setObjectName("gridLayout22")
        self.check_result_view = QtWidgets.QTreeView(self.gridLayoutWidget22)
        self.check_result_view.setAlternatingRowColors(True)
        self.check_result_view.setObjectName("check_result_view")
        self.gridLayout22.addWidget(self.check_result_view, 2, 0, 1, 1)
        self.check_result_search_box = SearchBox(self.gridLayoutWidget22)
        self.check_result_search_box.setObjectName("check_result_search_box")
        self.gridLayout22.addWidget(self.check_result_search_box, 0, 0, 1, 1)
        self.gridLayout_5.addWidget(self.splitter, 0, 0, 1, 1)
        MainWindow.setCentralWidget(self.centralwidget)
        self.menubar = QtWidgets.QMenuBar(MainWindow)
        self.menubar.setGeometry(QtCore.QRect(0, 0, 1200, 30))
        self.menubar.setObjectName("menubar")
        self.menu_File = QtWidgets.QMenu(self.menubar)
        self.menu_File.setObjectName("menu_File")
        self.action_new_value_set_2 = QtWidgets.QMenu(self.menu_File)
        self.action_new_value_set_2.setObjectName("action_new_value_set_2")
        self.menu_Project = QtWidgets.QMenu(self.menu_File)
        self.menu_Project.setObjectName("menu_Project")
        self.menuHelp = QtWidgets.QMenu(self.menubar)
        self.menuHelp.setObjectName("menuHelp")
        self.menuTools = QtWidgets.QMenu(self.menubar)
        self.menuTools.setObjectName("menuTools")
        MainWindow.setMenuBar(self.menubar)
        self.statusbar = QtWidgets.QStatusBar(MainWindow)
        self.statusbar.setObjectName("statusbar")
        MainWindow.setStatusBar(self.statusbar)
        self.action_open_project = QtWidgets.QAction(MainWindow)
        self.action_open_project.setObjectName("action_open_project")
        self.action_save_as_project = QtWidgets.QAction(MainWindow)
        self.action_save_as_project.setObjectName("action_save_as_project")
        self.action_save_project = QtWidgets.QAction(MainWindow)
        self.action_save_project.setObjectName("action_save_project")
        self.action_new_project = QtWidgets.QAction(MainWindow)
        self.action_new_project.setObjectName("action_new_project")
        self.action_about = QtWidgets.QAction(MainWindow)
        self.action_about.setObjectName("action_about")
        self.action_import_sym = QtWidgets.QAction(MainWindow)
        self.action_import_sym.setObjectName("action_import_sym")
        self.action_export_sym = QtWidgets.QAction(MainWindow)
        self.action_export_sym.setObjectName("action_export_sym")
        self.action_Project = QtWidgets.QAction(MainWindow)
        self.action_Project.setObjectName("action_Project")
        self.action_new_value_set = QtWidgets.QAction(MainWindow)
        self.action_new_value_set.setObjectName("action_new_value_set")
        self.action_Value_Set = QtWidgets.QAction(MainWindow)
        self.action_Value_Set.setObjectName("action_Value_Set")
        self.action_save_value_set = QtWidgets.QAction(MainWindow)
        self.action_save_value_set.setObjectName("action_save_value_set")
        self.action_save_as_value_set = QtWidgets.QAction(MainWindow)
        self.action_save_as_value_set.setObjectName("action_save_as_value_set")
        self.action_open_value_set = QtWidgets.QAction(MainWindow)
        self.action_open_value_set.setObjectName("action_open_value_set")
        self.action_import_smdx = QtWidgets.QAction(MainWindow)
        self.action_import_smdx.setObjectName("action_import_smdx")
        self.action_full_import = QtWidgets.QAction(MainWindow)
        self.action_full_import.setObjectName("action_full_import")
        self.action_full_export = QtWidgets.QAction(MainWindow)
        self.action_full_export.setObjectName("action_full_export")
        self.action_full_export_first_time = QtWidgets.QAction(MainWindow)
        self.action_full_export_first_time.setObjectName("action_full_export_first_time")
        self.action_check = QtWidgets.QAction(MainWindow)
        self.action_check.setObjectName("action_check")
        self.action_new_value_set_2.addAction(self.action_new_value_set)
        self.action_new_value_set_2.addAction(self.action_open_value_set)
        self.action_new_value_set_2.addAction(self.action_save_value_set)
        self.action_new_value_set_2.addAction(self.action_save_as_value_set)
        self.menu_Project.addAction(self.action_new_project)
        self.menu_Project.addAction(self.action_open_project)
        self.menu_Project.addAction(self.action_save_project)
        self.menu_Project.addAction(self.action_save_as_project)
        self.menu_File.addAction(self.menu_Project.menuAction())
        self.menu_File.addAction(self.action_new_value_set_2.menuAction())
        self.menu_File.addSeparator()
        self.menu_File.addAction(self.action_import_smdx)
        self.menu_File.addAction(self.action_full_export)
        self.menuHelp.addAction(self.action_about)
        self.menuTools.addAction(self.action_check)
        self.menubar.addAction(self.menu_File.menuAction())
        self.menubar.addAction(self.menuTools.menuAction())
        self.menubar.addAction(self.menuHelp.menuAction())

        self.retranslateUi(MainWindow)
        self.tabWidget.setCurrentIndex(0)
        QtCore.QMetaObject.connectSlotsByName(MainWindow)

    def retranslateUi(self, MainWindow):
        _translate = QtCore.QCoreApplication.translate
        MainWindow.setWindowTitle(_translate("MainWindow", "MainWindow"))
        self.tabWidget.setTabText(self.tabWidget.indexOf(self.tab_2), _translate("MainWindow", "CAN"))
        self.tabWidget.setTabText(self.tabWidget.indexOf(self.tab_21), _translate("MainWindow", "SunSpec"))
        self.tabWidget.setTabText(self.tabWidget.indexOf(self.tab), _translate("MainWindow", "Value Set"))
        self.menu_File.setTitle(_translate("MainWindow", "&File"))
        self.action_new_value_set_2.setTitle(_translate("MainWindow", "&Value Set"))
        self.menu_Project.setTitle(_translate("MainWindow", "&Project"))
        self.menuHelp.setTitle(_translate("MainWindow", "&Help"))
        self.menuTools.setTitle(_translate("MainWindow", "&Tools"))
        self.action_open_project.setText(_translate("MainWindow", "&Open..."))
        self.action_save_as_project.setText(_translate("MainWindow", "Save &As..."))
        self.action_save_project.setText(_translate("MainWindow", "&Save"))
        self.action_save_project.setShortcut(_translate("MainWindow", "Ctrl+S"))
        self.action_new_project.setText(_translate("MainWindow", "&New"))
        self.action_about.setText(_translate("MainWindow", "&About..."))
        self.action_import_sym.setText(_translate("MainWindow", "&Import .sym..."))
        self.action_export_sym.setText(_translate("MainWindow", "&Export .sym..."))
        self.action_Project.setText(_translate("MainWindow", "&Project"))
        self.action_new_value_set.setText(_translate("MainWindow", "&New"))
        self.action_Value_Set.setText(_translate("MainWindow", "&Value Set..."))
        self.action_save_value_set.setText(_translate("MainWindow", "&Save"))
        self.action_save_as_value_set.setText(_translate("MainWindow", "Save &As..."))
        self.action_open_value_set.setText(_translate("MainWindow", "&Open..."))
        self.action_import_smdx.setText(_translate("MainWindow", "Import &SMDX..."))
        self.action_full_import.setText(_translate("MainWindow", "Full import..."))
        self.action_full_export.setText(_translate("MainWindow", "&Full Export..."))
        self.action_full_export_first_time.setText(_translate("MainWindow", "Full Export (First Time)..."))
        self.action_check.setText(_translate("MainWindow", "&Check"))
from epyqlib.searchbox import SearchBox



import re

import attr
import toolz

import epyqlib.pm.parametermodel
import epyqlib.utils.general

import epcpm.c


builders = epyqlib.utils.general.TypeMap()


# TODO: get rid of this or get it somewhere else (pm exclude-from-SIL bool?)
def ignore_item(item):
    if 'sunspec' in item.variable.casefold():
        return True

    if 'unused placeholder' in item.variable:
        return True

    return False


@attr.s
class CHContents:
    c = attr.ib(factory=list)
    h = attr.ib(factory=list)

    def extend(self, other):
        self.c.extend(other.c)
        self.h.extend(other.h)

    def append(self, other):
        self.c.append(other.c)
        self.h.append(other.h)


def export(c_path, h_path, parameters_model):
    builder = builders.wrap(
        wrapped=parameters_model.root,
    )

    c_path.parent.mkdir(parents=True, exist_ok=True)

    built, items = builder.gen()

    template_context = {
        'item_count': len(items),
        'initializers': epcpm.c.format_nested_lists(built.c).rstrip(),
        'declarations': epcpm.c.format_nested_lists(built.h).rstrip(),
    }

    epcpm.c.render(
        source=c_path.with_suffix(f'{c_path.suffix}_pm'),
        destination=c_path,
        context=template_context,
    )

    epcpm.c.render(
        source=h_path.with_suffix(f'{h_path.suffix}_pm'),
        destination=h_path,
        context=template_context,
    )


def collect_items(parameters_root):
    all_items = []

    parameters = next(
        node
        for node in parameters_root.children
        if node.name == 'Parameters'
    )

    for child in parameters.children:
        if not isinstance(
                child,
                (
                        epyqlib.pm.parametermodel.Group,
                        epyqlib.pm.parametermodel.Parameter,
                        epyqlib.pm.parametermodel.Table,
                ),
        ):
            continue

        items = builders.wrap(
            wrapped=child,
            path=(),
            parameters_root=parameters_root,
            parameter_uuid_finder=parameters_root.model.node_from_uuid,
        ).gen()

        all_items.extend(items)

    return all_items


@builders(epyqlib.pm.parametermodel.Root)
@attr.s
class Root:
    wrapped = attr.ib()

    def gen(self):
        contents = CHContents()
        types = set()

        items = collect_items(parameters_root=self.wrapped)
        for index, item in enumerate(items):
            types.add(item.type)
            item_lines = item.create_initializer(index=index)
            contents.c.extend(item_lines)

        everything = CHContents()

        everything.c.extend([
                contents.c,
        ])

        everything.h.extend([
            # 'typedef enum SetterTypes {',
            # [
            #     f'setter_{type},'
            #     for type in types
            # ],
            # '} SetterTypes;',
            # '',
            # *[
            #     f'typedef void (*SetterPointer_{type})({type});'
            #     for type in types
            # ],
            # '',
            # 'typedef union Setter {',
            # [
            #     f'SetterPointer_{type} {type}_;'
            #     for type in types
            # ],
            # '} Setter;',
            # '',
            f'extern Item SIL_interfaceItems[{len(items)}];',
        ])

        return everything, items


@builders(epyqlib.pm.parametermodel.Group)
@attr.s
class Group:
    wrapped = attr.ib()
    path = attr.ib()
    parameters_root = attr.ib()
    parameter_uuid_finder = attr.ib()

    def gen(self):
        items = []

        for child in self.wrapped.children:
            if not isinstance(
                    child,
                    (
                        epyqlib.pm.parametermodel.Group,
                        epyqlib.pm.parametermodel.Parameter,
                        epyqlib.pm.parametermodel.Table,
                    ),
            ):
                continue

            built = builders.wrap(
                wrapped=child,
                path=self.path + (self.wrapped.name,),
                parameters_root=self.parameters_root,
                parameter_uuid_finder=self.parameter_uuid_finder,
            ).gen()

            items.extend(built)

        return items


@attr.s
class Item:
    uuid = attr.ib()
    variable = attr.ib()
    type = attr.ib()
    on_write = attr.ib()
    internal_scale = attr.ib()
    is_table = attr.ib(default=False)
    table_info = attr.ib(default=None)
    path = attr.ib(default=[])
    #     factory=lambda: TableInfo(zone=0, curve=0, index=0, setter=0, type=''),
    # )

    # TODO: index -> designator
    def create_initializer(self, index=None):
        initializers = self.create_subinitializers()

        if index is not None:
            index_text = f'[{index}] = '
        else:
            index_text = ''

        path = ' > '.join(self.path)
        path_comment = f' // {path}'

        item_initializer = [
            f'{index_text}{{{path_comment}',
            initializers,
            '},',
        ]

        return item_initializer

    def create_subinitializers(self):
        if self.table_info is None:
            table_info_initializer = []
        else:
            table_info_initializer = self.table_info.create_initializer(
                designator='.tableInfo',
            )

        is_table = 'true' if self.is_table else 'false'

        initializers = [
            f'.uuid = "{self.uuid}",',
            f'.setterType = setter_{self.type},',
            f'.setter = {{ .{self.type}_ = {self.on_write} }},',
            f'.variable = {{ .{self.type}_ = {self.variable} }},',
            f'.internalScale = {self.internal_scale},',
            f'.isTable = {is_table},',
            *table_info_initializer,
        ]

        return initializers


@attr.s
class TableInfo:
    zone = attr.ib()
    curve = attr.ib()
    index = attr.ib()
    setter = attr.ib()
    type = attr.ib()

    # TODO: index -> designator
    def create_initializer(self, designator=None):
        initializers = self.create_subinitializers()

        if designator is not None:
            designator_text = f'{designator} = '
        else:
            designator_text = ''

        item_initializer = [
            f'{designator_text}{{',
            initializers,
            '},',
        ]

        return item_initializer

    def create_subinitializers(self):
        initializers = []

        if self.zone is not None:
            initializers.append(f'.zone = {self.zone},')

        initializers.extend([
            f'.curve = {self.curve},',
            f'.index = {self.index},',
            # TODO: ugh, again with tables being all the same type internally
            f'.setter = {{ .int16_t_ = {self.setter} }},',
        ])

        return initializers


@builders(epyqlib.pm.parametermodel.Parameter)
@attr.s
class Parameter:
    wrapped = attr.ib()
    path = attr.ib()
    parameters_root = attr.ib()
    parameter_uuid_finder = attr.ib()

    def gen(self):
        parameter = self.wrapped

        supported_item_type = (
            parameter.uses_interface_item()
        )

        if not supported_item_type:
            return []

        if parameter.internal_type == 'void*':
            return []

        if 'txRate' in parameter.name:
            return []

        if parameter.setter_function is None:
            on_write = 'NULL'
        else:
            on_write = f'&{parameter.setter_function}'

        if parameter.internal_variable is None:
            variable = 'NULL'
        else:
            variable = f'&{parameter.internal_variable}'

        item = Item(
            uuid=parameter.uuid,
            variable=variable,
            type=parameter.internal_type,
            on_write=on_write,
            internal_scale=parameter.internal_scale_factor,
            path=self.path,
        )

        if ignore_item(item):
            return []

        return [item]


# TODO: CAMPid 68945967541316743769675426795146379678431
def breakdown_nested_array(s):
    split = re.split(r'\[(.*?)\].', s)

    array_layers = list(toolz.partition(2, split))
    remainder, = split[2 * len(array_layers):]

    return array_layers, remainder


# TODO: CAMPid 079549750417808543178043180
def get_curve_type(combination_string):
    # TODO: backmatching
    return {
        'LowRideThrough': 'IEEE1547_CURVE_TYPE_LRT',
        'HighRideThrough': 'IEEE1547_CURVE_TYPE_HRT',
        'LowTrip': 'IEEE1547_CURVE_TYPE_LTRIP',
        'HighTrip': 'IEEE1547_CURVE_TYPE_HTRIP',
    }.get(combination_string)


# TODO: CAMPid 0974567213671436714671907842679364
@attr.s
class NestedArrays:
    array_layers = attr.ib()
    remainder = attr.ib()

    @classmethod
    def build(cls, s):
        array_layers, remainder = breakdown_nested_array(s)

        return cls(
            array_layers=array_layers,
            remainder=remainder,
        )

    def index(self, indexes):
        try:
            return '.'.join(
                '{layer}[{index}]'.format(
                    layer=layer,
                    index=index_format.format(**indexes),
                )
                for (layer, index_format), index in zip(self.array_layers, indexes)
            )
        except KeyError as e:
            raise

    def sizeof(self, layers):
        indexed = self.index(indexes={layer: 0 for layer in layers})

        return f'sizeof({indexed})'

    def full(self, indexes):
        variable_base = self.index(indexes)
        variable = f'{variable_base}.{self.remainder}'

        return variable


# TODO: CAMPid 0795436754762451671643967431
# TODO: get this from the ...  wherever we have it
axes = ['x', 'y', 'z']


@builders(epyqlib.pm.parametermodel.Table)
@attr.s
class Table:
    wrapped = attr.ib()
    path = attr.ib()
    parameters_root = attr.ib()
    parameter_uuid_finder = attr.ib()

    def gen(self):
        group, = (
            child
            for child in self.wrapped.children
            if isinstance(child, epyqlib.pm.parametermodel.TableGroupElement)
        )

        arrays = [
            child
            for child in self.wrapped.children
            if isinstance(child, epyqlib.pm.parametermodel.Array)
        ]

        array_nests = {
            name: NestedArrays.build(s=array.children[0].internal_variable)
            for name, array in zip(axes, arrays)
        }

        items = builders.wrap(
            wrapped=group,
            path=self.path + (self.wrapped.name,),
            table=self.wrapped,
            array_nests=array_nests,
            parameter_uuid_finder=self.parameter_uuid_finder,
        ).gen()

        return items


@builders(epyqlib.pm.parametermodel.TableGroupElement)
@attr.s
class TableGroupElement:
    wrapped = attr.ib()
    path = attr.ib()
    table = attr.ib()
    array_nests = attr.ib()
    parameter_uuid_finder = attr.ib()
    layers = attr.ib(default=[])

    def gen(self):
        items = []

        table_tree_root = not isinstance(
            self.wrapped.tree_parent,
            epyqlib.pm.parametermodel.TableGroupElement,
        )

        layers = list(self.layers)
        if not table_tree_root:
            layers.append(self.wrapped.name)

        for child in self.wrapped.children:
            result = builders.wrap(
                wrapped=child,
                path=self.path + (self.wrapped.name,),
                table=self.table,
                layers=layers,
                array_nests=self.array_nests,
                parameter_uuid_finder=self.parameter_uuid_finder,
            ).gen()

            items.extend(result)

        return items


@builders(epyqlib.pm.parametermodel.TableArrayElement)
@attr.s
class TableArrayElement:
    wrapped = attr.ib()
    path = attr.ib()
    table = attr.ib()
    layers = attr.ib()
    array_nests = attr.ib()
    parameter_uuid_finder = attr.ib()

    def gen(self):
        table_element = self.wrapped

        # TODO: CAMPid 9655426754319431461354643167
        array_element = table_element.original

        if isinstance(array_element, epyqlib.pm.parametermodel.Parameter):
            parameter = array_element
        else:
            parameter = array_element.tree_parent.children[0]

        is_group = isinstance(
            parameter.tree_parent,
            epyqlib.pm.parametermodel.Group,
        )

        if is_group:
            return self.handle_group()

        return self.handle_array()

    def handle_array(self):
        table_element = self.wrapped

        # TODO: CAMPid 9655426754319431461354643167
        array_element = table_element.original

        if isinstance(array_element, epyqlib.pm.parametermodel.Parameter):
            parameter = array_element
        else:
            parameter = array_element.tree_parent.children[0]

        indexes = {
            'curve_type': get_curve_type(''.join(self.layers[:2])),
            'curve_index': int(self.layers[-2]) - 1,
            'point_index': int(table_element.name.lstrip('_').lstrip('0')) - 1,
        }

        axis = axes[self.table.arrays.index(parameter.tree_parent)]
        variable = self.array_nests[axis].full(indexes)
        # This cast covers the fact that all table points are internally
        # int16_t despite some being used as uint16_t.
        # TODO: verify compatible size at least?
        variable = f'({parameter.internal_type} *) &{variable}'

        if parameter.setter_function is None:
            # TODO: should Item do this?
            table_on_write = 'NULL'
        else:
            table_on_write = parameter.setter_function.format(upper_axis=axis.upper())

        table_info = TableInfo(
            zone=indexes['curve_type'],
            curve=indexes['curve_index'],
            index=indexes['point_index'],
            setter=table_on_write,
            type=parameter.internal_type,
        )

        return [
            Item(
                uuid=table_element.uuid,
                variable=variable,
                type=parameter.internal_type,
                on_write='NULL',
                internal_scale=parameter.internal_scale_factor,
                is_table=True,
                table_info=table_info,
                path=self.path,
            )
        ]

    def handle_group(self):
        table_element = self.wrapped

        curve_index = int(self.layers[-2]) - 1

        parameter = table_element.original

        if parameter.internal_type == 'PackedString':
            return []

        if parameter.internal_variable is None:
            return []

        if parameter.setter_function is None:
            # TODO: i think it's reasonable for Item to handle this?
            setter_function = 'NULL'
        else:
            setter_function = '&' + parameter.setter_function

        curve_type = get_curve_type(''.join(self.layers[:2]))

        internal_variable = parameter.internal_variable.format(
            curve_type=curve_type,
            curve_index=curve_index,
        )

        item = Item(
            uuid=table_element.uuid,
            variable=f'&{internal_variable}',
            type=parameter.internal_type,
            on_write=setter_function,
            internal_scale=parameter.internal_scale_factor,
            path=self.path,
        )

        return [item]



import collections
import io
import textwrap

import attr
import canmatrix.canmatrix
import canmatrix.formats

import epyqlib.pm.parametermodel
import epyqlib.utils.general

import epcpm.canmodel
import epcpm.symtoproject

builders = epyqlib.utils.general.TypeMap()


def dehumanize_name(name):
    return name
#     name = name.replace('-', '_')
#     return epyqlib.utils.general.spaced_to_upper_camel(name)


def export(path, can_model, parameters_model):
    finder = can_model.node_from_uuid
    access_levels = parameters_model.list_selection_roots['access level']
    builder = epcpm.cantosym.builders.wrap(
        wrapped=can_model.root,
        access_levels=access_levels,
        parameter_uuid_finder=finder,
        parameter_model=parameters_model,
    )

    path.parent.mkdir(parents=True, exist_ok=True)
    with path.open('w', newline='\n') as f:
        f.write(builder.gen())


class SignalOutsideMessageError(Exception):
    @classmethod
    def build(cls, signal, message_length):
        path = [signal]

        while True:
            parent = path[0].tree_parent

            if parent.tree_parent is None:
                break

            path.insert(0, parent)

        path_string = ' : '.join(element.name for element in path)

        message_range = [0, max(0, (message_length * 8) - 1)]
        signal_range = [signal.start_bit, signal.start_bit + signal.bits - 1]

        message = (
              f'{path_string} spans bits'
              f' [{signal_range[0]}, {signal_range[1]}]'
              f' which is outside the range'
              f' [{message_range[0]}, {message_range[1]}]'
        )

        return cls(message)


@builders(epcpm.canmodel.Root)
@attr.s
class Root:
    wrapped = attr.ib()
    access_levels = attr.ib()
    parameter_uuid_finder = attr.ib(default=None)
    parameter_model = attr.ib(default=None)

    def gen(self):
        matrix = canmatrix.canmatrix.CanMatrix()
        # TODO: this shouldn't need to be copied from:
        #           canmatrix.sym.load()
        matrix.add_frame_defines("GenMsgCycleTime", 'INT 0 65535')
        matrix.add_frame_defines("Receivable", 'BOOL False True')
        matrix.add_frame_defines("Sendable", 'BOOL False True')
        matrix.add_signal_defines("GenSigStartValue", 'FLOAT -3.4E+038 3.4E+038')
        matrix.add_signal_defines("HexadecimalOutput", 'BOOL False True')
        matrix.add_signal_defines("DisplayDecimalPlaces", 'INT 0 65535')
        matrix.add_signal_defines("LongName", 'STR')

        for child in self.wrapped.children:
            frame = builders.wrap(
                wrapped=child,
                access_levels=self.access_levels,
                parameter_uuid_finder=self.parameter_uuid_finder,
            ).gen()
            matrix.add_frame(frame)

        enumerations = self.collect_enumerations()
        used_enumerations = {
            signal.enumeration
            for frame in matrix.frames
            for signal in frame.signals
            if signal.enumeration is not None
        }

        for enumeration in enumerations:
            if enumeration.name not in used_enumerations:
                continue

            enumerators = collections.OrderedDict(
                (e.value, dehumanize_name(e.name))
                for e in enumeration.children
            )

            matrix.add_value_table(
                name=dehumanize_name(enumeration.name),
                valueTable=enumerators,
            )

        codec = 'utf-8'

        f = io.BytesIO()
        canmatrix.formats.dump(matrix, f, 'sym', symExportEncoding=codec)
        f.seek(0)

        return f.read().decode(codec)

    def collect_enumerations(self):
        collected = []

        if self.parameter_model is None:
            return collected

        def collect(node, collected):
            is_enumeration = isinstance(
                node,
                (
                    epyqlib.pm.parametermodel.Enumeration,
                    epyqlib.pm.parametermodel.AccessLevels,
                )
            )
            if is_enumeration:
                collected.append(node)

        self.parameter_model.root.traverse(
            call_this=collect,
            payload=collected,
            internal_nodes=True,
        )

        return collected


@builders(epcpm.canmodel.Message)
@attr.s
class Message:
    wrapped = attr.ib()
    access_levels = attr.ib()
    parameter_uuid_finder = attr.ib(default=None)

    def gen(self):
        frame = canmatrix.canmatrix.Frame(
            name=dehumanize_name(self.wrapped.name),
            arbitration_id=canmatrix.canmatrix.ArbitrationId(
                id=self.wrapped.identifier,
                extended=self.wrapped.extended,
            ),
            size=self.wrapped.length,
            comment=self.wrapped.comment,
            cycle_time=(
                self.wrapped.cycle_time
                if self.wrapped.cycle_time is not None
                else 0
            ),
        )

        frame.attributes['Receivable'] = str(self.wrapped.receivable)
        frame.attributes['Sendable'] = str(self.wrapped.sendable)

        for child in self.wrapped.children:
            signal = builders.wrap(
                wrapped=child,
                message_length=self.wrapped.length,
                parameter_uuid_finder=self.parameter_uuid_finder,
            ).gen()
            frame.signals.append(signal)

        return frame


@builders(epcpm.canmodel.Signal)
@attr.s
class Signal:
    wrapped = attr.ib()
    message_length = attr.ib()
    parameter_uuid_finder = attr.ib(default=None)

    def gen(
            self,
            multiplex_id=None,
            multiplex_on_write=None,
            skip_access_level=False,
            skip_configuration=False,
    ):
        if (
                self.message_length is not None
                and (
                    self.wrapped.start_bit < 0
                    or (
                        self.message_length * 8
                        < self.wrapped.start_bit + self.wrapped.bits
                    )
                )
        ):
            raise SignalOutsideMessageError.build(
                signal=self.wrapped,
                message_length=self.message_length,
            )

        extras = {}
        can_find_parameter = (
            self.wrapped.parameter_uuid is not None
            and self.parameter_uuid_finder is not None
        )
        parameter = None
        initial_value = None
        if can_find_parameter:
            parameter = self.parameter_uuid_finder(self.wrapped.parameter_uuid)

            original_parameter = parameter
            if isinstance(
                    parameter,
                    epyqlib.pm.parametermodel.TableArrayElement,
            ):
                array_element = parameter.original

                if isinstance(
                        array_element,
                        epyqlib.pm.parametermodel.Parameter,
                ):
                    original_parameter = array_element
                else:
                    original_parameter = array_element.tree_parent.children[0]

            if parameter.minimum is not None:
                extras['min'] = parameter.minimum

            if parameter.maximum is not None:
                extras['max'] = parameter.maximum

            if parameter.comment is not None:
                comment = parameter.comment.strip()
                if len(comment) > 0:
                    extras['comment'] = comment

            extras['comment'] = '{comment} <rw:{r}:{w}>'.format(
                comment=extras.get('comment', ''),
                r=1,
                w=0 if parameter.read_only else 1,
            ).strip()

            handle_access_level = (
                    not skip_access_level
                    and parameter.access_level_uuid is not None
            )
            if handle_access_level:
                access_level = self.parameter_uuid_finder(
                    parameter.access_level_uuid
                )
                if access_level != access_level.tree_parent.default():
                    extras['comment'] = '{} <{}>'.format(
                        extras.get('comment', ''),
                        access_level.name.casefold(),
                    ).strip()

            handle_configuration = (
                    not skip_configuration
                    and parameter.visibility is not None
            )
            if handle_configuration:
                configurations = [
                    self.parameter_uuid_finder(u) for u in parameter.visibility
                ]
                imported_variants = epcpm.symtoproject.imported_variants
                if all(v in configurations  for v in imported_variants):
                    configurations = None #don't spam sym with variants
                if configurations is not None:
                    for cfg in configurations:
                        if cfg is not None:
                            extras['comment'] = '{} <{}>'.format(
                                extras.get('comment', ''),
                                cfg.name,
                            ).strip()

            if parameter.nv_format is not None:
                segments = ['nv']

                nv_flags = ''
                if parameter.nv_cast:
                    nv_flags += 'c'

                segments.append(nv_flags)

                if parameter.nv_factor is not None:
                    segments.append('f{}'.format(parameter.nv_factor))

                segments.append(parameter.nv_format)

                extras['comment'] = '{}  <{}>'.format(
                    extras.get('comment', ''),
                    ':'.join(segments),
                ).strip()
            elif (
                    isinstance(
                        original_parameter,
                        epyqlib.pm.parametermodel.Parameter,
                    )
                    and original_parameter.uses_interface_item()
            ):
                is_table = False
                ancestor = original_parameter.tree_parent
                while ancestor is not None:
                    if isinstance(ancestor, epyqlib.pm.parametermodel.Table):
                        is_table = True
                        break
                    ancestor = ancestor.tree_parent

                if is_table:
                    getter = 'table_items_getMeta'
                    setter = 'table_items_setMeta'
                else:
                    getter = 'items_getMeta'
                    setter = 'items_setMeta'

                if multiplex_on_write is None:
                    multiplex_on_write = ''

                comment_format_interface_item_segments = ':'.join([
                    'InterfaceItem',
                    '{item}',
                    '{getter}',
                    '{setter}',
                    '{multiplex_on_write}'
                ])
                comment_format = (
                    f'{{comment}}  <{comment_format_interface_item_segments}>'
                )
                extras['comment'] = comment_format.format(
                    comment=extras.get('comment', ''),
                    item='interfaceItem_{}'.format(
                        str(self.wrapped.parameter_uuid).replace('-', '_'),
                    ),
                    getter=getter,
                    setter=setter,
                    multiplex_on_write=multiplex_on_write,
                ).strip()

            comment = extras.get('comment', '')
            extras['comment'] = f'{comment}  <uuid:{parameter.uuid}>'.strip()

            if parameter.units is not None:
                extras['unit'] = parameter.units

            if self.wrapped.enumeration_uuid is not None:
                enumeration = self.parameter_uuid_finder(
                    self.wrapped.enumeration_uuid,
                )

                extras['enumeration'] = dehumanize_name(enumeration.name)
                extras['values'] = {v: k for k, v in enumeration.items()}

            if parameter.default is not None:
                initial_value = parameter.default

        if initial_value is not None:
            extras['initial_value'] = initial_value

        signal = canmatrix.canmatrix.Signal(
            name=dehumanize_name(self.wrapped.name),
            multiplex=multiplex_id,
            size=self.wrapped.bits,
            is_signed=self.wrapped.signed,
            factor=self.wrapped.factor,
            start_bit=self.wrapped.start_bit,
            calc_min_for_none=False,
            calc_max_for_none=False,
            **extras,
        )

        if parameter is not None:
            attributes = signal.attributes

            attributes['LongName'] = parameter.name
            attributes['HexadecimalOutput'] = parameter.display_hexadecimal

            if parameter.decimal_places is not None:
                attributes['DisplayDecimalPlaces'] = parameter.decimal_places

        return signal


@builders(epcpm.canmodel.MultiplexedMessage)
@attr.s
class MultiplexedMessage:
    wrapped = attr.ib()
    access_levels = attr.ib()
    parameter_uuid_finder = attr.ib(default=None)

    def gen(self):
        common_signals = []
        not_signals = []
        table_multiplexers = set()
        for child in self.wrapped.children[1:]:
            if isinstance(child, epcpm.canmodel.Signal):
                common_signals.append(child)
            elif isinstance(child, epcpm.canmodel.CanTable):
                for subchild in child.children:
                    if isinstance(subchild, epcpm.canmodel.Multiplexer):
                        not_signals.append(subchild)
                        table_multiplexers.add(subchild)
            else:
                not_signals.append(child)

        frame = canmatrix.canmatrix.Frame(
            name=dehumanize_name(self.wrapped.name),
            arbitration_id=canmatrix.canmatrix.ArbitrationId(
                id=self.wrapped.identifier,
                extended=self.wrapped.extended,
            ),
            size=not_signals[0].length,
            comment=self.wrapped.comment,
            cycle_time=(
                not_signals[0].cycle_time
                if not_signals[0].cycle_time is not None
                else 0
            ),
            attributes={
                'Receivable': str(self.wrapped.receivable),
                'Sendable': str(self.wrapped.sendable),
            },
        )

        if len(self.wrapped.children) == 0:
            return frame

        mux_signal = builders.wrap(
            wrapped=self.wrapped.children[0],
            message_length=None,
            parameter_uuid_finder=self.parameter_uuid_finder,
        ).gen(
            multiplex_id='Multiplexor',
        )
        frame.signals.append(mux_signal)

        for multiplexer in not_signals:
            if multiplexer.comment is not None:
                mux_signal.comments[multiplexer.identifier] = (
                    multiplexer.comment
                )

            # TODO: backmatching
            if multiplexer in table_multiplexers:
                mux_signal.comments[multiplexer.identifier] = (
                    '{} <{}>'.format(
                        mux_signal.comments.get(multiplexer.identifier, ''),
                        'table',
                    ).strip()
                )

            name = multiplexer.name
            if isinstance(multiplexer.tree_parent, epcpm.canmodel.CanTable):
                name = multiplexer.tree_parent.name + name
            frame.mux_names[multiplexer.identifier] = (
                dehumanize_name(name)
            )

            def param_special(signal):
                folded = signal.name.casefold()

                return folded.startswith('read param - ') or folded == 'meta'

            signal_access_levels = set()

            for signal in multiplexer.children:
                if param_special(signal):
                    continue

                parameter = self.parameter_uuid_finder(signal.parameter_uuid)
                uuid = parameter.access_level_uuid

                if uuid is None:
                    access_level = self.access_levels.default()
                else:
                    access_level = self.parameter_uuid_finder(uuid)

                signal_access_levels.add(access_level)

            all_access_levels_match = len(signal_access_levels) == 1

            if all_access_levels_match:
                access_level = signal_access_levels.pop()
                if access_level != access_level.tree_parent.default():
                    mux_signal.comments[multiplexer.identifier] = (
                        '{} <{}>'.format(
                            mux_signal.comments.get(multiplexer.identifier, ''),
                            access_level.name.casefold(),
                        ).strip()
                    )

            for signal in multiplexer.children:
                signal = builders.wrap(
                    wrapped=signal,
                    message_length=multiplexer.length,
                    parameter_uuid_finder=self.parameter_uuid_finder,
                ).gen(
                    multiplex_id=multiplexer.identifier,
                    multiplex_on_write=multiplexer.on_write,
                    skip_access_level=all_access_levels_match,
                )

                frame.signals.append(signal)

            frame_signal_names = [
                signal.name
                for signal in frame.signals
                if signal.multiplex == multiplexer.identifier
            ]

            for signal in reversed(common_signals):
                if signal.name in frame_signal_names:
                    continue

                matrix_signal = builders.wrap(
                    wrapped=signal,
                    message_length=multiplexer.length,
                    parameter_uuid_finder=self.parameter_uuid_finder,
                ).gen(
                    multiplex_id=multiplexer.identifier,
                )
                frame.signals.insert(0, matrix_signal)

        return frame


def tweak_reply_signal(sig):
    if sig.name.endswith('_command'):
        sig.name = sig.name.replace('_command', '_status')
        sig.attributes['LongName'] = sig.name
    return sig


@builders(epcpm.canmodel.MultiplexedMessageClone)
@attr.s
class MultiplexedMessageClone:
    wrapped = attr.ib()
    access_levels = attr.ib()
    parameter_uuid_finder = attr.ib(default=None)

    def gen(self):
        frame = builders.wrap(
            wrapped=self.wrapped.original,
            access_levels=self.access_levels,
            parameter_uuid_finder=self.parameter_uuid_finder,
        ).gen()
 
        frame.name = self.wrapped.name
        frame.arbitration_id = canmatrix.canmatrix.ArbitrationId(
            id=self.wrapped.identifier,
            # TODO: should technically have it's own extended attribute
            #       rather than grabbing from the original
            extended=self.wrapped.original.extended,
        )
        frame.comment = self.wrapped.comment
        frame.attributes={
            'Receivable': str(self.wrapped.receivable),
            'Sendable': str(self.wrapped.sendable),
        }
        for sig in frame.signals[:]:
            sig = tweak_reply_signal(sig)

        return frame





import textwrap

import graham

import epcpm.project

import pathlib

import epyqlib.pm

reference_string = textwrap.dedent('''\
{
    "_type": "project",
    "paths": {
        "_type": "models",
        "parameters": "parameters.json",
        "can": "can.json",
        "sunspec": "sunspec.json"
    }
}''')

reference_project = epcpm.project.Project()
reference_project.paths.parameters = 'parameters.json'
reference_project.paths.can = 'can.json'
reference_project.paths.sunspec = 'sunspec.json'


def test_save():
    assert reference_string == graham.dumps(reference_project, indent=4).data


def test_load():
    assert reference_project == (
        graham.schema(epcpm.project.Project).loads(reference_string).data
    )


def test_model_iterable():
    model = epcpm.project.Models()

    assert tuple(model) == ('parameters', 'can', 'sunspec')


def test_model_set_all():
    model = epcpm.project.Models()

    value = object()

    model.set_all(value)

    assert all(v is value for v in model.values())


def test_model_items():
    model = epcpm.project.Models()

    expected = (
        ('parameters', None),
        ('can', None),
        ('sunspec', None),
    )

    assert tuple(model.items()) == expected


def test_model_values():
    model = epcpm.project.Models()

    for name in model:
        model[name] = name + '_'

    assert tuple(model.values()) == ('parameters_', 'can_', 'sunspec_')


def test_model_getitem():
    model = epcpm.project.Models()

    values = []
    for name in model:
        values.append(model[name])

    assert values == [None, None, None]


def test_model_proper_selection_roots():
    project = epcpm.project.loadp(
        pathlib.Path(__file__).with_name('example_project.pmp')
    )

    expected = epyqlib.pm.parametermodel.types.list_selection_roots()
    
    assert set(project.models.parameters.list_selection_roots.keys()) == expected



import itertools
import functools

import attr
import pycparser

import epyqlib.pm.parametermodel
import epyqlib.utils.general


builders = epyqlib.utils.general.TypeMap()


@builders(epyqlib.pm.parametermodel.Root)
@attr.s
class Root:
    wrapped = attr.ib()

    def definition(self):
        definitions = []

        for member in self.wrapped.children:
            builder = builders.wrap(member)
            definitions.extend(builder.definition())

        return definitions

    def instantiation(self):
        decls = []

        for member in self.wrapped.children:
            builder = builders.wrap(member)

            decls.append(Decl(
                type=Type(
                    name=epyqlib.utils.general.spaced_to_lower_camel(
                        member.name,
                    ),
                    type=builder.type_name(),
                )
            ))

        return decls


@builders(epyqlib.pm.parametermodel.Parameter)
@attr.s
class Parameter:
    wrapped = attr.ib()

    def definition(self):
        return []

    def type_name(self):
        type_name = self.wrapped.type_name

        if type_name is None:
            return 'void'

        return type_name


@builders(epyqlib.pm.parametermodel.Group)
@attr.s
class Group:
    wrapped = attr.ib()

    def definition(self):
        definitions = []
        member_decls = []

        for member in self.wrapped.children:
            builder = builders.wrap(member)

            member_decls.append(Decl(
                type=Type(
                    name=epyqlib.utils.general.spaced_to_lower_camel(
                        member.name,
                    ),
                    type=builder.type_name(),
                )
            ))

            definitions.extend(builder.definition())

        return [
            *definitions,
            *struct(
                name=self.type_name()[:-2],
                member_decls=member_decls,
            ),
        ]

    def type_name(self):
        name = self.wrapped.type_name
        if name is None:
            name = self.wrapped.name
        return epyqlib.utils.general.spaced_to_upper_camel(name) + '_t'


@builders(epyqlib.pm.parametermodel.Array)
@attr.s
class Array:
    wrapped = attr.ib()

    def definition(self):
        builder = builders.wrap(self.wrapped.children[0])
        definitions = builder.definition()

        values = []
        if self.wrapped.named_enumerators:
            values.extend(enumerate(
                epyqlib.utils.general.spaced_to_upper_camel(child.name)
                for child in self.wrapped.children
            ))

        values.append((len(self.wrapped.children), 'Count'))

        enum_definitions = enum(
            name=self.base_type_name(),
            enumerators=[
                (
                    '{base}_{name}'.format(
                        base=self.base_type_name(),
                        name=name,
                    ),
                    value,
                )
                for value, name in values
            ],
        )

        return [
            *definitions,
            *enum_definitions,
            array_typedef(
                target=builder.type_name(),
                name=self.type_name(),
                length=pycparser.c_ast.ID(
                    enum_definitions[0].type.values.enumerators[-1].name
                ),
            ),
        ]

    def base_type_name(self):
        return epyqlib.utils.general.spaced_to_upper_camel(self.wrapped.name)

    def type_name(self):
        return self.base_type_name() + '_t'


def int_literal(value):
    return pycparser.c_ast.Constant(type='int', value=str(value))


def Type(name, type):
    return pycparser.c_ast.TypeDecl(
        declname=name,
        quals=[],
        type=pycparser.c_ast.IdentifierType(
            names=(type,),
        ),
    )


Decl = functools.partial(
    pycparser.c_ast.Decl,
    name=None,
    quals=[],
    storage=[],
    funcspec=[],
    init=None,
    bitsize=None,
)

ArrayDecl = functools.partial(
    pycparser.c_ast.ArrayDecl,
    dim_quals=[],
)

TypeDecl = functools.partial(
    pycparser.c_ast.TypeDecl,
    declname='',
    quals=[],
)


def typedef(target, name):
    return pycparser.c_ast.Typedef(
        name='',
        quals=[],
        storage=['typedef'],
        type=pycparser.c_ast.TypeDecl(
            declname=name,
            quals=[],
            type=target,
        ),
    )


def array_typedef(target, name, length):
    return pycparser.c_ast.Typedef(
        name=name,
        quals=[],
        storage=['typedef'],
        type=ArrayDecl(
            dim=length,
            type=TypeDecl(
                declname=name,
                type=pycparser.c_ast.IdentifierType(
                    names=[target],
                ),
            ),
        ),
    )


def enum(name, enumerators=()):
    enum_name = '{name}_e'.format(name=name)
    typedef_name = '{name}_et'.format(name=name)

    enumerators = pycparser.c_ast.EnumeratorList(enumerators=tuple(
        pycparser.c_ast.Enumerator(name=name, value=int_literal(value))
        for name, value in enumerators
    ))

    enumeration = pycparser.c_ast.Enum(
        name=enum_name,
        values=enumerators
    )

    declaration = Decl(type=enumeration)

    typedef_ = typedef(
        target=pycparser.c_ast.Enum(
            name=enum_name,
            values=None
        ),
        name=typedef_name,
    )

    return declaration, typedef_


def struct(name, member_decls=()):
    struct_name = f'{name}_s'
    typedef_name = f'{name}_t'

    struct = pycparser.c_ast.Struct(
        name=struct_name,
        decls=member_decls,
    )

    decl = pycparser.c_ast.Decl(
        name=None,
        quals=[],
        storage=[],
        funcspec=[],
        type=struct,
        init=None,
        bitsize=None,
    )

    typedef_ = typedef(
        target=pycparser.c_ast.Struct(
            name=struct_name,
            decls=[],
        ),
        name=typedef_name,
    )

    return decl, typedef_


def array(type, name, length):
    return Decl(
        name=name,
        type=ArrayDecl(
            type=Type(name, type),
            dim=int_literal(length),
        )
    )



import textwrap

import epcpm.c


def test_format_nested_lists():
    example = [
        'void getSUNSPEC_MODEL17_Nam(void) {',
        [
            'size_t i;',
            'UartName name = modbusHandlerGetName()',
            '',
            'for (i = 0; i < LENGTHOF(name.s); i++) {',
            [
                'sunspecInterface.model17.Nam[i] = name.s[i];',
            ],
            '}',
        ],
        '}',
    ]

    result = epcpm.c.format_nested_lists(it=example)

    assert result == textwrap.dedent('''\
    void getSUNSPEC_MODEL17_Nam(void) {
        size_t i;
        UartName name = modbusHandlerGetName()

        for (i = 0; i < LENGTHOF(name.s); i++) {
            sunspecInterface.model17.Nam[i] = name.s[i];
        }
    }
    ''')
